# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.10.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
import sys
import logging
from .ttypes import *
from thrift.Thrift import TProcessor
from thrift.transport import TTransport


class Iface(object):
    """
    Base service that all other services should inherit from
    """
    def about(self):
        """
        Get information about the service
        """
        pass

    def alive(self):
        """
        Is the service alive?
        """
        pass


class Client(Iface):
    """
    Base service that all other services should inherit from
    """
    def __init__(self, iprot, oprot=None):
        self._iprot = self._oprot = iprot
        if oprot is not None:
            self._oprot = oprot
        self._seqid = 0

    def about(self):
        """
        Get information about the service
        """
        self.send_about()
        return self.recv_about()

    def send_about(self):
        self._oprot.writeMessageBegin('about', TMessageType.CALL, self._seqid)
        args = about_args()
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_about(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = about_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "about failed: unknown result")

    def alive(self):
        """
        Is the service alive?
        """
        self.send_alive()
        return self.recv_alive()

    def send_alive(self):
        self._oprot.writeMessageBegin('alive', TMessageType.CALL, self._seqid)
        args = alive_args()
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_alive(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = alive_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "alive failed: unknown result")


class Processor(Iface, TProcessor):
    def __init__(self, handler):
        self._handler = handler
        self._processMap = {}
        self._processMap["about"] = Processor.process_about
        self._processMap["alive"] = Processor.process_alive

    def process(self, iprot, oprot):
        (name, type, seqid) = iprot.readMessageBegin()
        if name not in self._processMap:
            iprot.skip(TType.STRUCT)
            iprot.readMessageEnd()
            x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
            oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
            x.write(oprot)
            oprot.writeMessageEnd()
            oprot.trans.flush()
            return
        else:
            self._processMap[name](self, seqid, iprot, oprot)
        return True

    def process_about(self, seqid, iprot, oprot):
        args = about_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = about_result()
        try:
            result.success = self._handler.about()
            msg_type = TMessageType.REPLY
        except (TTransport.TTransportException, KeyboardInterrupt, SystemExit):
            raise
        except Exception as ex:
            msg_type = TMessageType.EXCEPTION
            logging.exception(ex)
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("about", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_alive(self, seqid, iprot, oprot):
        args = alive_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = alive_result()
        try:
            result.success = self._handler.alive()
            msg_type = TMessageType.REPLY
        except (TTransport.TTransportException, KeyboardInterrupt, SystemExit):
            raise
        except Exception as ex:
            msg_type = TMessageType.EXCEPTION
            logging.exception(ex)
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("alive", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

# HELPER FUNCTIONS AND STRUCTURES


class about_args(object):

    thrift_spec = (
    )

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('about_args')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class about_result(object):
    """
    Attributes:
     - success
    """

    thrift_spec = (
        (0, TType.STRUCT, 'success', (ServiceInfo, ServiceInfo.thrift_spec), None, ),  # 0
    )

    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = ServiceInfo()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('about_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class alive_args(object):

    thrift_spec = (
    )

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('alive_args')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class alive_result(object):
    """
    Attributes:
     - success
    """

    thrift_spec = (
        (0, TType.BOOL, 'success', None, None, ),  # 0
    )

    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.BOOL:
                    self.success = iprot.readBool()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('alive_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.BOOL, 0)
            oprot.writeBool(self.success)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)



"""Code for reading and writing Concrete Communications
"""
from __future__ import unicode_literals

try:
    from cStringIO import StringIO as BytesIO
except ImportError:
    from io import BytesIO

import gzip
import bz2
import mimetypes
import tarfile
import zipfile

import os
import time

from thrift import TSerialization
from thrift.transport import TTransport

from ..communication.ttypes import Communication
from ..structure.ttypes import TokenLattice
from .references import add_references_to_communication
from .thrift_factory import factory


if os.name == 'nt':
    def _get_uid():
        return 0

    def _get_gid():
        return 0

    def _get_username():
        return ''

    def _get_groupname():
        return ''

else:
    import pwd
    import grp

    def _get_uid():
        return os.getuid()

    def _get_gid():
        return os.getgid()

    def _get_username():
        return pwd.getpwuid(_get_uid()).pw_name

    def _get_groupname():
        return grp.getgrgid(_get_gid()).gr_name


def read_thrift_from_file(thrift_obj, filename):
    """Instantiate Thrift object from contents of named file

    The Thrift file is assumed to be encoded using TCompactProtocol

    **WARNING** - Thrift deserialization tends to fail silently.  For
    example, the Thrift libraries will not complain if you try to
    deserialize data from the file `/dev/urandom`.

    Args:
        thrift_obj: A Thrift object (e.g. a Communication object)
        filename (str):  A filename string

    Returns:
        The Thrift object that was passed in as an argument
    """
    thrift_file = open(filename, "rb")
    thrift_bytes = thrift_file.read()
    TSerialization.deserialize(
        thrift_obj, thrift_bytes,
        protocol_factory=factory.protocolFactory)
    thrift_file.close()
    return thrift_obj


def read_communication_from_file(communication_filename, add_references=True):
    """Read a Communication from the file specified by filename

    Args:
        communication_filename (str): String with filename
        add_references (bool): If True, calls
           :func:`concrete.util.references.add_references_to_communication`
           on :class:`.Communication` read from file

    Returns:
        Communication:
    """
    comm = read_thrift_from_file(Communication(), communication_filename)
    if add_references:
        add_references_to_communication(comm)
    return comm


def read_tokenlattice_from_file(tokenlattice_filename):
    """Read a :class:`.TokenLattice` from a file

    Args:
        tokenlattice_filename (str): Name of file containing serialized
            :class:`.TokenLattice`

    Returns:
        TokenLattice:
    """
    return read_thrift_from_file(TokenLattice(), tokenlattice_filename)


def write_communication_to_file(communication, communication_filename):
    """Write a :class:`.Communication` to a file

    Args:
        communication (Communication)
        communication_filename (str)
    """
    return write_thrift_to_file(communication, communication_filename)


def write_thrift_to_file(thrift_obj, filename):
    """Write a Thrift object to a file

    Args:
        thrift_obj:
        filename (str):
    """
    thrift_bytes = TSerialization.serialize(
        thrift_obj,
        protocol_factory=factory.protocolFactory)
    thrift_file = open(filename, "wb")
    thrift_file.write(thrift_bytes)
    thrift_file.close()


class _FileTypeClass(object):
    '''
    An instance of this class represents filetypes abstractly; its
    members correspond to individual filetypes.  It probably doesn't
    make sense to have more than one instance of this class.
    '''

    def __init__(self, *names):
        self.CHOICES = tuple(names)
        for (i, name) in enumerate(names):
            if name == 'CHOICES':
                raise ValueError('%s is an invalid filetype name' % name)
            setattr(self, self._normalize(name), i)

    @classmethod
    def _normalize(cls, name):
        return name.replace('-', '_').upper()

    def lookup(self, ft):
        '''
        Return filetype (integer value) for ft, where ft may be
        a filetype name or (for convenience) the filetype integer
        value itself.
        '''

        if isinstance(ft, int):
            return ft
        elif isinstance(ft, str) or isinstance(ft, unicode):
            return getattr(self, self._normalize(ft))
        else:
            raise ValueError('unknown filetype %s' % str(ft))

    def add_argument(self, argparse_parser, flag='--input-filetype',
                     help_name='input file'):
        '''
        Add filetype argument to parser, return function that can be
        called on argparse namespace to retrieve argument value, e.g.,

        >>> get_filetype = add_argument(parser, ...)
        >>> ns = parser.parse_args()
        >>> filetype = get_filetype(ns)
        '''
        argparse_parser.add_argument(flag, type=str, choices=self.CHOICES,
                                     default=self.CHOICES[0],
                                     help='filetype for %s (choices: %s)' %
                                          (help_name, ', '.join(self.CHOICES)))

        def _get_value(ns):
            return getattr(ns, flag.lstrip('-').replace('-', '_'))
        return _get_value


FileType = _FileTypeClass(
    'auto',
    'zip',
    'tar', 'tar-gz', 'tar-bz2',
    'stream', 'stream-gz', 'stream-bz2',
)


class ThriftReader(object):
    """Iterator/generator class for reading one or more Thrift structures
    from a file

    The iterator returns a `(obj, filename)` tuple where obj is an object
    of type thrift_type.

    Supported filetypes are:

    - a file with a single Thrift structure
    - a file with multiple Thrift structures concatenated together
    - a gzipped file with a single Thrift structure
    - a gzipped file with multiple Thrift structures concatenated
      together
    - a .tar.gz file with one or more Thrift structures
    - a .zip file with one or more Thrift structures

    Sample usage::

        for (comm, filename) in ThriftReader(Communication,
                                             'multiple_comms.tar.gz'):
            do_something(comm)
    """

    def __init__(self, thrift_type, filename,
                 postprocess=None, filetype=FileType.AUTO):
        """
        Args:
            thrift_type: Class for Thrift type, e.g. Communication, TokenLattice
            filename (str):
            postprocess (function): A post-processing function that is called
                with the Thrift object as argument each time a Thrift object
                is read from the file
            filetype (FileType): Expected type of file.  Default value is
                `FileType.AUTO`, where function will try to automatically
                determine file type.
        """
        filetype = FileType.lookup(filetype)

        self._thrift_type = thrift_type
        if postprocess is None:
            def _noop(obj):
                return
            self._postprocess = _noop
        else:
            self._postprocess = postprocess
        self._source_filename = filename

        if filetype == FileType.TAR:
            self.filetype = 'tar'
            self.tar = tarfile.open(filename, 'r|')

        elif filetype == FileType.TAR_GZ:
            self.filetype = 'tar'
            self.tar = tarfile.open(filename, 'r|gz')

        elif filetype == FileType.TAR_BZ2:
            self.filetype = 'tar'
            self.tar = tarfile.open(filename, 'r|bz2')

        elif filetype == FileType.ZIP:
            self.filetype = 'zip'
            self.zip = zipfile.ZipFile(filename, 'r')
            self.zip_infolist = self.zip.infolist()
            self.zip_infolist_index = 0

        elif filetype == FileType.STREAM:
            self.filetype = 'stream'
            f = open(filename, 'rb')

        elif filetype == FileType.STREAM_GZ:
            self.filetype = 'stream'
            f = gzip.open(filename, 'rb')

        elif filetype == FileType.STREAM_BZ2:
            self.filetype = 'stream'
            f = bz2.BZ2File(filename, 'r')

        elif filetype == FileType.AUTO:
            if tarfile.is_tarfile(filename):
                self.filetype = 'tar'
                self.tar = tarfile.open(filename, 'r|*')

            elif zipfile.is_zipfile(filename):
                self.filetype = 'zip'
                self.zip = zipfile.ZipFile(filename, 'r')
                self.zip_infolist = self.zip.infolist()
                self.zip_infolist_index = 0

            elif mimetypes.guess_type(filename)[1] == 'gzip':
                # this is not a true stream---is_tarfile will have
                # successfully seeked backwards on the file if we have
                # reached this point
                self.filetype = 'stream'
                f = gzip.open(filename, 'rb')

            elif mimetypes.guess_type(filename)[1] == 'bzip2':
                # this is not a true stream
                self.filetype = 'stream'
                f = bz2.BZ2File(filename, 'r')

            else:
                # this is not a true stream
                self.filetype = 'stream'
                f = open(filename, 'rb')

        else:
            raise ValueError('unknown filetype %d' % filetype)

        if self.filetype == 'stream':
            self.transport = TTransport.TFileObjectTransport(f)
            self.protocol = factory.createProtocol(self.transport)
            self.transport.open()

    def __iter__(self):
        return self

    def __next__(self):
        """Returns a `(obj, filename)` tuple where obj is an object
        of type thrift_type.

        If the ThriftReader is reading from an archive, then
        `filename` will be set to the name of the Thrift entry in
        the archive (e.g. `foo.concrete`), and not the name of the
        archive file (e.g. `bar.zip`).  If the ThriftReader is reading
        from a concatenated file (instead of an archive), then all
        Thrift structures extracted from the concatenated file will have
        the same value for the `filename` field.
        """
        if self.filetype == 'stream':
            return self._next_from_stream()
        elif self.filetype == 'tar':
            return self._next_from_tar()
        elif self.filetype == 'zip':
            return self._next_from_zip()
        else:
            raise ValueError('unknown filetype %s' % self.filetype)

    def next(self):
        return self.__next__()

    def _next_from_stream(self):
        try:
            comm = self._thrift_type()
            comm.read(self.protocol)
            self._postprocess(comm)
            return (comm, self._source_filename)
        except EOFError:
            self.transport.close()
            raise StopIteration

    def _next_from_tar(self):
        while True:
            tarinfo = self.tar.next()
            if tarinfo is None:
                raise StopIteration
            if not tarinfo.isfile():
                # Ignore directories
                continue
            filename = os.path.split(tarinfo.name)[-1]
            if filename[0] == '.' and filename[1] == '_':
                # Ignore attribute files created by OS X tar
                continue
            comm = TSerialization.deserialize(
                self._thrift_type(),
                self.tar.extractfile(tarinfo).read(),
                protocol_factory=factory.protocolFactory)
            self._postprocess(comm)
            # hack to keep memory usage O(1)
            # (...but the real hack is tarfile :)
            self.tar.members = []
            return (comm, tarinfo.name)

    def _next_from_zip(self):
        if self.zip_infolist_index >= len(self.zip_infolist):
            raise StopIteration
        zipinfo = self.zip_infolist[self.zip_infolist_index]
        self.zip_infolist_index += 1
        comm = TSerialization.deserialize(
            self._thrift_type(),
            self.zip.open(zipinfo).read(),
            protocol_factory=factory.protocolFactory)
        self._postprocess(comm)
        return (comm, zipinfo.filename)


class CommunicationReader(ThriftReader):
    """Iterator/generator class for reading one or more Communications from a
    file

    The iterator returns a `(Communication, filename)` tuple

    Supported filetypes are:

    - a file with a single Communication
    - a file with multiple Communications concatenated together
    - a gzipped file with a single Communication
    - a gzipped file with multiple Communications concatenated together
    - a .tar.gz file with one or more Communications
    - a .zip file with one or more Communications

    Sample usage::

        for (comm, filename) in CommunicationReader('multiple_comms.tar.gz'):
            do_something(comm)
    """

    def __init__(self, filename, add_references=True, filetype=FileType.AUTO):
        """
        Args:
            filename (str):
            add_references (bool): If True, calls
               :func:`concrete.util.references.add_references_to_communication`
               on all :class:`.Communication` objects read from file
            filetype (FileType): Expected type of file.  Default value is
                `FileType.AUTO`, where function will try to automatically
                determine file type.
        """
        super(CommunicationReader, self).__init__(
            Communication,
            filename,
            postprocess=(add_references_to_communication
                         if add_references
                         else None),
            filetype=filetype)


class CommunicationWriter(object):
    """Class for writing one or more Communications to a file

    Sample usage::

        writer = CommunicationWriter('foo.concrete')
        writer.write(existing_comm_object)
        writer.close()
    """

    def __init__(self, filename=None):
        if filename is not None:
            self.open(filename)

    def close(self):
        self.file.close()

    def open(self, filename):
        """
        Args:
            filename (str)
        """
        self.file = open(filename, 'wb')

    def write(self, comm):
        """
        Args:
            comm (Communication)
        """
        thrift_bytes = TSerialization.serialize(
            comm, protocol_factory=factory.protocolFactory)
        self.file.write(thrift_bytes)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


class CommunicationWriterTar(object):
    """Class for writing one or more Communications to a .TAR archive

    Sample usage::

        writer = CommunicationWriterTar('multiple_comms.tar')
        writer.write(comm_object_one, 'comm_one.concrete')
        writer.write(comm_object_two, 'comm_two.concrete')
        writer.write(comm_object_three, 'comm_three.concrete')
        writer.close()
    """

    def __init__(self, tar_filename=None, gzip=False):
        """
        Args:
            tar_filename(str): If a filename is given, :func:`open`
                will be called with the filename
            gzip (bool): Flag indicating if .TAR file should be
                compressed with gzip
        """
        self.gzip = gzip
        if tar_filename is not None:
            self.open(tar_filename)

    def close(self):
        self.tarfile.close()

    def open(self, tar_filename):
        """
        Args:
            tar_filename (str):
        """
        self.tarfile = tarfile.open(tar_filename, 'w:gz' if self.gzip else 'w')

    def write(self, comm, comm_filename=None):
        """
        Args:
            comm (Communication)
            comm_filename (str)
        """
        if comm_filename is None:
            comm_filename = comm.id + '.concrete'

        thrift_bytes = TSerialization.serialize(
            comm, protocol_factory=factory.protocolFactory)

        file_like_obj = BytesIO(thrift_bytes)

        comm_tarinfo = tarfile.TarInfo()
        comm_tarinfo.type = tarfile.REGTYPE
        comm_tarinfo.name = comm_filename
        comm_tarinfo.size = len(thrift_bytes)
        comm_tarinfo.mode = 0o644
        comm_tarinfo.mtime = time.time()
        comm_tarinfo.uid = _get_uid()
        comm_tarinfo.uname = _get_username()
        comm_tarinfo.gid = _get_gid()
        comm_tarinfo.gname = _get_groupname()

        self.tarfile.addfile(comm_tarinfo, file_like_obj)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


class CommunicationWriterTGZ(CommunicationWriterTar):
    """Class for writing one or more Communications to a .TAR.GZ archive

    Sample usage::

        writer = CommunicationWriterTGZ('multiple_comms.tgz')
        writer.write(comm_object_one, 'comm_one.concrete')
        writer.write(comm_object_two, 'comm_two.concrete')
        writer.write(comm_object_three, 'comm_three.concrete')
        writer.close()
    """

    def __init__(self, tar_filename=None):
        super(CommunicationWriterTGZ, self).__init__(tar_filename, gzip=True)



#!/usr/bin/env python

"""Command line client for FetchCommunicationService
"""
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import sys

import concrete.version
from concrete.access import FetchCommunicationService
from concrete.access.ttypes import FetchRequest
from concrete.util.file_io import CommunicationWriterTGZ
from concrete.util.thrift_factory import factory
from concrete.util import set_stdout_encoding


def main():
    set_stdout_encoding()

    parser = argparse.ArgumentParser(
        description="Command line client for interacting with a FetchCommunicationService server",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("-p", "--port", type=int, default=9090,
                        help="Port of FetchCommunicationService server")
    parser.add_argument("-s", "--server", default="localhost",
                        help="Hostname of FetchCommunicationService server")
    parser.add_argument("--about", action="store_true",
                        help="Print value of fetch_service.about()")
    parser.add_argument("--alive", action="store_true",
                        help="Print value of fetch_service.alive()")
    parser.add_argument("--count", action="store_true",
                        help="Print value of fetch_service.getCommunicationCount()")
    parser.add_argument("--get-ids", action="store_true",
                        help="Print list of Communication IDs returned by "
                        "fetch_service.getCommunicationIDs(offset, count).  "
                        "The offset and count parameters are set using the "
                        "'--get-ids-offset' and '--get-ids-count' flags")
    parser.add_argument("--get-ids-offset", type=int, default=0, metavar="ID_OFFSET",
                        help="Number of Communication IDs printed using the '--get-ids' flag")
    parser.add_argument("--get-ids-count", type=int, default=20, metavar="ID_COUNT",
                        help="Offset for Communication IDs printed using the '--get-ids' flag")
    parser.add_argument("--save-as-tgz", metavar="TGZ_FILENAME",
                        help="Save fetched Communications to a TGZ archive containing files "
                        "named '[COMMUNICATION_ID].concrete'")
    parser.add_argument("comm_id", nargs="*", help="IDs of Communications to be fetched. "
                        "If '-' is specified, a list of Communication IDs will be read from "
                        "stdin, one Communication ID per line")
    concrete.version.add_argparse_argument(parser)
    args = parser.parse_args()

    socket = factory.createSocket(args.server, args.port)
    transport = factory.createTransport(socket)
    protocol = factory.createProtocol(transport)
    client = FetchCommunicationService.Client(protocol)
    transport.open()

    if args.comm_id:
        fetch_request = FetchRequest()
        if len(args.comm_id) == 1 and args.comm_id[0] == '-':
            fetch_request.communicationIds = [line.strip() for line in sys.stdin.readlines()]
        else:
            fetch_request.communicationIds = args.comm_id
        fetch_result = client.fetch(fetch_request)
        print("Received FetchResult: '%s'" % fetch_result)

    if args.about:
        print("FetchCommunicationService.about() returned %s" % client.about())
    if args.alive:
        print("FetchCommunicationService.alive() returned %s" % client.alive())
    if args.count:
        print("FetchCommunicationService.getCommunicationCount() returned %d" %
              client.getCommunicationCount())
    if args.get_ids:
        print("FetchCommunicationService.getCommunicationIDs(offset=%d, count=%d) returned:" %
              (args.get_ids_offset, args.get_ids_count))
        for comm_id in client.getCommunicationIDs(args.get_ids_offset, args.get_ids_count):
            print("  %s" % comm_id)

    if args.save_as_tgz and args.comm_id:
        if fetch_result.communications:
            with CommunicationWriterTGZ(args.save_as_tgz) as writer:
                for comm in fetch_result.communications:
                    comm_filename = '%s.concrete' % comm.id
                    print("Saving Communication to TGZ archive '%s' as '%s'" %
                          (args.save_as_tgz, comm_filename))
                    writer.write(comm, comm_filename)


if __name__ == '__main__':
    main()



from __future__ import unicode_literals
from multiprocessing import Process
from time import sleep
from socket import create_connection

from .thrift_factory import factory


class ConcreteServiceClientWrapper(object):
    def __init__(self, host, port):
        if not hasattr(self, 'concrete_service_class'):
            raise NotImplementedError(
                "Child classes of ConcreteServiceClientWrapper must set " +
                "the 'concrete_service_class' attribute to a class that " +
                "implements a Concrete Service")

        self.host = host
        try:
            self.port = int(port)
        except ValueError:
            raise ValueError(
                "Service expected 'port' to be an integer, but it was '%s'" %
                port)

    def __enter__(self):
        socket = factory.createSocket(self.host, self.port)
        self.transport = factory.createTransport(socket)
        protocol = factory.createProtocol(self.transport)

        cli = self.concrete_service_class.Client(protocol)

        self.transport.open()
        return cli

    def __exit__(self, type, value, traceback):
        self.transport.close()


class ConcreteServiceWrapper(object):
    """
    A sample wrapper around a Concrete service.
    """

    def __init__(self, implementation):
        if not hasattr(self, 'concrete_service_class'):
            raise NotImplementedError(
                "Child classes of ConcreteServiceWrapper must set " +
                "the 'concrete_service_class' attribute to a class that " +
                "implements a Concrete Service")

        self.processor = self.concrete_service_class.Processor(implementation)

    def serve(self, host, port):
        server = factory.createServer(self.processor, host, port)

        # NOTE: Thrift's servers run indefinitely. This server implementation
        # may be killed by a KeyboardInterrupt (Control-C); otherwise, the
        # process must be killed to terminate the server.
        server.serve()


class SubprocessConcreteServiceWrapper(object):
    """
    Concrete Service wrapper that runs server in a subprocess via a
    context manager interface.
    """

    SLEEP_INTERVAL = 0.1

    def __init__(self, implementation, host, port, timeout=None):
        if not hasattr(self, 'concrete_service_wrapper_class'):
            raise NotImplementedError(
                "Child classes of SubprocessConcreteServiceWrapper must " +
                "set the 'concrete_service_wrapper_class' attribute to a " +
                "child class of ConcreteServiceWrapper")

        self.proc = None
        self.server = self.concrete_service_wrapper_class(implementation)
        self.host = host
        self.port = port
        self.timeout = timeout

    def __enter__(self):
        self.proc = Process(target=self.server.serve,
                            args=(self.host, self.port))
        self.proc.start()

        s = None
        elapsed = 0.

        # Loop until successful connect or timeout
        while s is None and (self.timeout is None or elapsed < self.timeout):
            try:
                s = create_connection((self.host, self.port))
                s.close()
            except:
                s = None

                # this is not precise... not important
                sleep(self.SLEEP_INTERVAL)
                elapsed += self.SLEEP_INTERVAL

        return self

    def __exit__(self, type, value, traceback):
        self.proc.terminate()
        self.proc.join()



from __future__ import unicode_literals
import logging
import os

from ..access.ttypes import FetchResult
from ..services.ttypes import ServiceInfo
from .access_wrapper import FetchCommunicationClientWrapper
from .file_io import write_communication_to_file
from ..version import concrete_library_version


class CommunicationContainerFetchHandler(object):
    """FetchCommunicationService implementation using Communication containers

    Implements the :mod:`.FetchCommunicationService` interface, retrieving
    Communications from a dict-like `communication_container` object
    that maps Communication ID strings to Communications.  The
    `communication_container` could be an actual dict, or a container
    such as:

    - :class:`.DirectoryBackedCommunicationContainer`
    - :class:`.FetchBackedCommunicationContainer`
    - :class:`.MemoryBackedCommunicationContainer`
    - :class:`.RedisHashBackedCommunicationContainer`
    - :class:`.ZipFileBackedCommunicationContainer`

    Usage::

        from concrete.util.access_wrapper import FetchCommunicationServiceWrapper

        handler = CommunicationContainerFetchHandler(comm_container)
        fetch_service = FetchCommunicationServiceWrapper(handler)
        fetch_service.serve(host, port)
    """

    def __init__(self, communication_container):
        """
        Args:
            communication_container: Dict-like object that maps Communication
                                     IDs to Communications
        """
        self.communication_container = communication_container

    def about(self):
        logging.info("Received about() call")
        service_info = ServiceInfo()
        service_info.name = 'CommunicationContainerFetchHandler - %s' % \
                            type(self.communication_container)
        service_info.version = concrete_library_version()
        return service_info

    def alive(self):
        logging.info("Received alive() call")
        return True

    def fetch(self, fetch_request):
        logging.info("Received FetchRequest: %s" % fetch_request)
        fetch_result = FetchResult()
        fetch_result.communications = []
        for communication_id in fetch_request.communicationIds:
            if communication_id in self.communication_container:
                comm = self.communication_container[communication_id]
                fetch_result.communications.append(comm)
            else:
                logging.warning('Unable to find Communication with ID: %s' % communication_id)
        return fetch_result

    def getCommunicationCount(self):
        logging.info('Received getCommunicationCount()')
        communicationCount = len(self.communication_container)
        logging.info('- Communication Count: %d' % communicationCount)
        return communicationCount

    def getCommunicationIDs(self, offset, count):
        logging.info('Received getCommunicationIDs() call')
        return list(self.communication_container.keys())[offset:][:count]


class DirectoryBackedStoreHandler(object):
    """Simple StoreCommunicationService implementation using a directory

    Implements the :mod:`.StoreCommunicationService` interface, storing
    Communications in a directory.
    """
    def __init__(self, store_path):
        """
        Args:
            store_path: Path where Communications should be Stored
        """
        self.store_path = store_path

    def about(self):
        logging.info("DirectoryBackedStoreHandler.about() called")
        service_info = ServiceInfo()
        service_info.name = 'DirectoryBackedStoreHandler'
        service_info.version = concrete_library_version()
        return service_info

    def alive(self):
        logging.info("DirectoryBackedStoreHandler.alive() called")
        return True

    def store(self, communication):
        """Save Communication to a directory

        Stored Communication files will be named `[COMMUNICATION_ID].comm`.
        If a file with that name already exists, it will be overwritten.
        """
        logging.info(
            "DirectoryBackedStoreHandler.store() called with Communication "
            "with ID '%s'" % communication.id)
        comm_filename = os.path.join(self.store_path, communication.id + '.comm')
        write_communication_to_file(communication, comm_filename)


class RelayFetchHandler(object):
    """Implements a 'relay' to another :mod:`.FetchCommunicationService`
    server.

    A :mod:`.FetchCommunicationService` that acts as a relay to a
    second :mod:`.FetchCommunicationService`, where the second
    service is using the TSocket transport and TCompactProtocol
    protocol.

    This class was designed for the use case where you have Thrift
    JavaScript code that needs to communicate with a
    :mod:`.FetchCommunicationService` server, but the server does
    not support the same Thrift serialization protocol as the
    JavaScript client.

    The de-facto standard for Concrete services is to use the
    TCompactProtocol serialization protocol over a TSocket connection.
    But as of Thrift 0.10.0, the Thrift JavaScript libraries only
    support using TJSONProtocol over HTTP.

    The RelayFetchHandler class is intended to be used as server-side
    code by a web application.  The JavaScript code will make
    :mod:`.FetchCommunicationService` RPC calls to the web server
    using HTTP/TJSONProtocol, and the web application will then pass
    these RPC calls to another :mod:`.FetchCommunicationService`
    using TSocket/TCompactProtocol RPC calls.
    """
    def __init__(self, host, port):
        """
        Args:
            host (str): Hostname of :mod:`.FetchCommunicationService` server
            port (int): Port # of :mod:`.FetchCommunicationService` server
        """
        self.host = host
        self.port = int(port)

    def about(self):
        logging.info('RelayFetchHandler.about()')
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.about()

    def alive(self):
        logging.info('RelayFetchHandler.alive()')
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.alive()

    def fetch(self, request):
        logging.info('RelayFetchHandler.fetch()')
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.fetch(request)

    def getCommunicationCount(self):
        logging.info('RelayFetchHandler.getCommunicationCount()')
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.getCommunicationCount()

    def getCommunicationIDs(self, offset, count):
        logging.info('RelayFetchHandler.getCommunicationIDs(offset=%d, count=%d)' %
                     (offset, count))
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.getCommunicationIDs(offset, count)



from __future__ import unicode_literals
from concrete.util import (
    get_index_of_tool, datetime_to_timestamp, now_timestamp
)

from concrete import AnnotationMetadata

from datetime import datetime
import random
import string


class HasMetadata(object):
    def __init__(self, tool=None):
        self.metadata = AnnotationMetadata(
            tool=HasMetadata.gen_tool(tool),
            timestamp=0,
        )

    @classmethod
    def gen_tool(cls, tool=None, size=6,
                 chars=string.ascii_uppercase + string.digits):
        return tool if (tool is not None) else \
            u''.join(random.choice(chars) for _ in range(size))


def test_get_index_of_tool_none_list():
    assert get_index_of_tool(None, None) == -1
    assert get_index_of_tool(None, "") == -1
    assert get_index_of_tool(None, "My awesome tool") == -1


def test_get_index_of_tool_empty_list():
    assert get_index_of_tool([], None) == -1
    assert get_index_of_tool([], "") == -1
    assert get_index_of_tool([], "My awesome tool") == -1


def test_get_index_of_tool_nonempty_list_contains():
    lst = [
        HasMetadata("My awesome tool"),
        HasMetadata("My awesome tool--new"),
        HasMetadata()
    ]
    assert get_index_of_tool(lst, None) == 0
    assert get_index_of_tool(lst, "") == -1
    assert get_index_of_tool(lst, "My awesome tool") == 0
    lst[2] = lst[0]
    lst[0] = HasMetadata()
    assert get_index_of_tool(lst, None) == 0
    assert get_index_of_tool(lst, "") == -1
    assert get_index_of_tool(lst, "My awesome tool") == 2


def test_get_index_of_tool_nonempty_list_no_contains():
    lst = [
        HasMetadata(),
        HasMetadata(),
        HasMetadata()
    ]
    assert get_index_of_tool(lst, None) == 0
    assert get_index_of_tool(lst, "") == -1
    assert get_index_of_tool(lst, "My awesome tool") == -1


def test_datetime_to_timestamp():
    assert datetime_to_timestamp(datetime(1970, 1, 1)) == 0
    assert datetime_to_timestamp(datetime(1970, 1, 1, 0, 0, 47)) == 47


def test_now_timestamp():
    assert now_timestamp() > datetime_to_timestamp(datetime(2000, 1, 1, 0, 0, 0))



"""Communication Containers - mapping Communication IDs to Communications

Classes that behave like a read-only dictionary (implementing Python's
`collections.Mapping` interface) and map Communication ID strings to
Communications.

The classes abstract away the storage backend.  If you need to
optimize for performance, you may not want to use a dictionary
abstraction that retrieves one Communication at a time.
"""
from __future__ import unicode_literals

import collections
import gzip
import logging
import os
import zipfile

import humanfriendly

from ..access.ttypes import FetchRequest
from .access_wrapper import FetchCommunicationClientWrapper
from .file_io import (
    CommunicationReader,
    read_communication_from_file)
from .mem_io import read_communication_from_buffer


class DirectoryBackedCommunicationContainer(collections.Mapping):
    """Maps Comm IDs to Comms, retrieving Comms from the filesystem

    `DirectoryBackedCommunicationContainer` instances behave as
    dict-like data structures that map Communication IDs to
    Communications.  Communications are lazily-retrieved from the
    filesystem.

    Upon initialization, a `DirectoryBackedCommunicationContainer`
    instance will (recursively) search `directory_path` for any files
    that end with the specified `comm_extensions`.  Files with
    matching extensions are assumed to be Communication files whose
    filename (sans extension) is the file's Communication ID.  So, for
    example, a file named 'XIN_ENG_20101212.0120.concrete' is assumed
    to be a Communication file with a Communication ID of
    'XIN_ENG_20101212.0120'.

    Files with the extension `.gz` will be decompressed using gzip.

    A `DirectoryBackedCommunicationsContainer` will not be able to find
    any files that are added to `directory_path` after the container
    was initialized.
    """

    def __init__(self, directory_path,
                 comm_extensions=['.comm', '.concrete', '.gz']):
        """
        Args:
             directory_path (str): Path to directory containing Communications files
             comm_extensions (str[]): List of strings specifying filename extensions
                                         to be associated with Communications
        """
        self.comm_id_to_comm_path = {}

        logging.info("Caching names of files with extensions [%s] in '%s'" %
                     (', '.join(comm_extensions), directory_path))

        # Recursively traverse subdirectories
        for root, dirs, files in os.walk(directory_path):
            for basename in files:
                (comm_id, extension) = os.path.splitext(basename)
                if extension in comm_extensions:
                    comm_filename = os.path.join(root, basename)
                    self.comm_id_to_comm_path[comm_id] = comm_filename

        logging.info("Finished caching %d Communication filenames in '%s'" %
                     (len(self.comm_id_to_comm_path), directory_path))

    def __getitem__(self, communication_id):
        if communication_id in self.comm_id_to_comm_path:
            comm_path = self.comm_id_to_comm_path[communication_id]
            logging.info("Reading Communication with ID '%s'" % comm_path)
            if not os.path.exists(comm_path):
                logging.error("Unable to find file with path '%s'" % comm_path)
                raise KeyError
            if os.path.splitext(comm_path)[1] == '.gz':
                with gzip.open(comm_path) as gzip_file:
                    buf = gzip_file.read()
                    comm = read_communication_from_buffer(buf)
            else:
                comm = read_communication_from_file(comm_path)
            return comm
        else:
            logging.info('No Communication with ID: %s' % communication_id)
            raise KeyError

    def __iter__(self):
        return self.comm_id_to_comm_path.__iter__()

    def __len__(self):
        return len(self.comm_id_to_comm_path)


class FetchBackedCommunicationContainer(collections.Mapping):
    """Maps Comm IDs to Comms, retrieving Comms from a
    :mod:`.FetchCommunicationService` server

    `FetchBackedCommunicationContainer` instances behave as dict-like data
    structures that map Communication IDs to Communications.  Communications
    are lazily-retrieved from a :mod:`.FetchCommunicationService`.

    If you need to retrieve large amounts of data from a
    :mod:`.FetchCommunicationService`, then you SHOULD NOT USE THIS CLASS.
    This class retrieves one Communication at a time using
    :mod:`.FetchCommunicationService`.
    """

    def __init__(self, host, port):
        """
        Args:
            host (str): Hostname of :mod:`.FetchCommunicationService` server
            port (int): Port # of :mod:`.FetchCommunicationService` server
        """
        self.host = host
        self.port = port

    def __getitem__(self, communication_id):
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            fetch_result = fc.fetch(FetchRequest(communicationIds=[communication_id]))
            total_results = len(fetch_result.communications)
            if total_results == 0:
                raise KeyError
            elif total_results == 1:
                return fetch_result.communications[0]
            else:
                raise Exception("FetchBackedCommunicationContainer.__get_item__() "
                                "expected to receive 1 Communication, but instead "
                                "received %d Communications" % total_results)

    def __iter__(self):
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            n = fc.getCommunicationCount()
            return fc.getCommunicationIDs(0, n).__iter__()

    def __len__(self):
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.getCommunicationCount()


class MemoryBackedCommunicationContainer(collections.Mapping):
    """Maps Comm IDs to Comms by loading all Comms in file into memory

    `FetchBackedCommunicationContainer` instances behave as dict-like
    data structures that map Communication IDs to Communications.  All
    Communications in `communications_file` will be read into memory
    using a :class:`.CommunicationReader` instance.
    """

    def __init__(self, communications_file, max_file_size=1073741824):
        """
        Args:
            communications_file (str): String specifying name of Communications file
            max_file_size (int): Maximum file size, in bytes
        """
        self.comm_id_to_comm = {}

        comm_file_size = os.path.getsize(communications_file)
        if comm_file_size > max_file_size:
            raise Exception(
                ("MemoryBackedComunicationContainer will not open the file "
                 "'%s' because the file's size (%s) is larger than the "
                 "maximum specified file size of %s.  If you would like to "
                 "read a file this large into memory, you will need to "
                 "specify a larger value for the 'max_file_size'parameter.") %
                (communications_file,
                 humanfriendly.format_size(comm_file_size, binary=True),
                 humanfriendly.format_size(max_file_size, binary=True)))

        logging.info("Reading in Communications from file '%s'" %
                     communications_file)
        logging.info("Communication IDs:")
        for (comm, _) in CommunicationReader(communications_file):
            self.comm_id_to_comm[comm.id] = comm
            logging.info("  %s" % comm.id)
        logging.info("Finished reading communications.\n")

    def __getitem__(self, communication_id):
        return self.comm_id_to_comm[communication_id]

    def __iter__(self):
        return self.comm_id_to_comm.__iter__()

    def __len__(self):
        return len(self.comm_id_to_comm)


class ZipFileBackedCommunicationContainer(collections.Mapping):
    """Maps Comm IDs to Comms, retrieving Comms from a Zip file

    `ZipFileBackedCommunicationContainer` instances behave as dict-like
    data structures that map Communication IDs to Communications.
    Communications are lazily-retrieved from a Zip file.
    """

    def __init__(self, zipfile_path, comm_extensions=['.comm', '.concrete']):
        """
        Args:
            zipfile_path (str): Path to Zip file containing Communications
            comm_extensions (str[]): List of strings specifying filename extensions
                                 associated with Communications
        """
        self.comm_id_to_filename = {}
        self.zipfile = zipfile.ZipFile(zipfile_path, 'r')
        for filename in self.zipfile.namelist():
            (comm_id, extension) = os.path.splitext(os.path.basename(filename))
            if extension in comm_extensions:
                self.comm_id_to_filename[comm_id] = filename

    def __exit__(self):
        self.zipfile.close()

    def __getitem__(self, communication_id):
        filename = self.comm_id_to_filename[communication_id]
        buf = self.zipfile.read(filename)
        comm = read_communication_from_buffer(buf)
        return comm

    def __iter__(self):
        return self.comm_id_to_filename.__iter__()

    def __len__(self):
        return len(self.comm_id_to_filename)


class RedisHashBackedCommunicationContainer(collections.Mapping):
    """Maps Comm IDs to Comms, retrieving Comms from a Redis hash

    `RedisHashBackedCommunicationContainer` instances behave as dict-like
    data structures that map Communication IDs to Communications.
    Communications are lazily-retrieved from a Redis hash.
    """

    def __init__(self, redis_db, key):
        """
        Args:
            redis_db (redis.Redis): redis database connection
            key (str): Key in redis database where hash is located
        """
        self.redis_db = redis_db
        self.key = key

    def __getitem__(self, communication_id):
        buf = self.redis_db.hget(self.key, communication_id)
        if buf is None:
            raise KeyError
        comm = read_communication_from_buffer(buf)
        return comm

    def __contains__(self, communication_id):
        return self.redis_db.hexists(self.key, communication_id)

    def __iter__(self):
        return iter(self.redis_db.hkeys(self.key))

    def __len__(self):
        return self.redis_db.hlen(self.key)



from __future__ import unicode_literals

import sys
import codecs


def set_stdout_encoding():
    if sys.version_info[0] < 3:
        sys.stdout = codecs.getwriter('utf-8')(sys.stdout)
    else:
        sys.stdout = codecs.getwriter('utf-8')(sys.stdout.detach())



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

import requests

from concrete.search.ttypes import SearchQuery, SearchType
from concrete.util import set_stdout_encoding
from concrete.util.search_wrapper import SearchClientWrapper
import concrete.version


def print_search_result(result, http_lookup_url):
    for result_item in result.searchResultItems:
        if http_lookup_url:
            print(requests.get(
                http_lookup_url %
                result_item.communicationId
            ).text)
        else:
            print(result_item.communicationId)


def main():
    set_stdout_encoding()

    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        description='Interface with a Concrete Search service'
    )
    parser.add_argument('host',
                        help='Hostname of search service to which to'
                             ' connect.')
    parser.add_argument('port', type=int,
                        help='Port of search service to which to connect.')
    parser.add_argument('--http-lookup-url', type=str,
                        help='Look up result communication text from HTTP '
                             'service via provided URL template, for '
                             'example, http://localhost:3000/comm/id/%%s')
    parser.add_argument('--user-id', type=str,
                        help='user id to send to search service')
    parser.add_argument("--k", type=int, default=10,
                        help="Maximum number of search results to return")
    parser.add_argument("--about", action="store_true",
                        help="Print output of searchService.about()")
    parser.add_argument("--alive", action="store_true",
                        help="Print output of searchService.alive()")
    parser.add_argument("--capabilities", action="store_true",
                        help="Print output of searchService.getCapabilities()")
    parser.add_argument("--corpora", action="store_true",
                        help="Print output of searchService.getCorpora()")
    parser.add_argument("terms", nargs="*")
    concrete.version.add_argparse_argument(parser)
    ns = parser.parse_args()

    with SearchClientWrapper(ns.host, ns.port) as client:
        interactive_mode = True

        if ns.about or ns.alive or ns.capabilities or ns.corpora or ns.terms:
            interactive_mode = False

        if ns.about:
            print("SearchService.about() returned %s" % client.about())
        if ns.alive:
            print("SearchService.alive() returned %s" % client.alive())
        if ns.capabilities:
            print("SearchService.getCapabilities() returned %s" % client.getCapabilities())
        if ns.corpora:
            print("SearchService.getCorpora() returned %s" % client.getCorpora())

        if interactive_mode:
            while True:
                try:
                    line = raw_input('> ').strip().decode('utf-8')
                except EOFError:
                    print()
                    break
                if line:
                    terms = line.split()
                    query = SearchQuery(k=ns.k,
                                        rawQuery=line,
                                        terms=terms,
                                        type=SearchType.COMMUNICATIONS,
                                        userId=ns.user_id)
                    print_search_result(client.search(query), ns.http_lookup_url)
        elif ns.terms:
            query = SearchQuery(k=ns.k,
                                rawQuery=' '.join(ns.terms),
                                terms=ns.terms,
                                type=SearchType.COMMUNICATIONS,
                                userId=ns.user_id)
            print_search_result(client.search(query), None)


if __name__ == '__main__':
    main()



"""Functions used by `concrete_inspect.py` to print data in a Communication.

The function implementations provide useful examples of how to
interact with many different Concrete datastructures.
"""
from __future__ import print_function
from __future__ import unicode_literals

from .util.metadata import get_index_of_tool
from .util.unnone import lun
from .util.tokenization import (
    get_tokenizations, get_tagged_tokens, NoSuchTokenTagging,
)
from collections import defaultdict
from operator import attrgetter

try:
    unicode
except NameError:
    unicode = str


def _reconcile_index_and_tool(lst_of_conc, given_idx, tool):
    """Given a list of Concrete objects with metadata (e.g., `DependencyParse`s)
    and a default index, find the index of the object whose `.metadata.tool`
    matches the provided query tool name `tool`.

    When no tool is provided (but with an iterable list), this returns the
    provided default index (even if it\'s not a valid index).
    If the tool isn't found, the list is None or empty, this returns -1.
    """
    valid_lst = lst_of_conc is not None and len(lst_of_conc) > 0
    idx = given_idx if valid_lst else -1
    if tool is not None:
        idx = get_index_of_tool(lst_of_conc, tool)
    return idx


def _valid_index_lun(lst, idx):
    """Return True iff `idx` is a valid index into
    the given (non-None) list. If `lst` is None,
    return False.
    """

    if lst is None or len(lst) == 0:
        return False
    return idx >= 0 and idx < len(lst)


def _filter_by_tool(lst, tool):
    return filter(lambda x: tool is None or x.metadata.tool == tool, lst)


def _get_tagged_token_strs_by_token_index(tagged_tokens, num_tokens):
    tagged_tokens_by_token_index = dict(
        (tagged_token.tokenIndex, tagged_token)
        for tagged_token in tagged_tokens
    )
    return [
        (
            tagged_tokens_by_token_index[token_index].tag
            if token_index in tagged_tokens_by_token_index
            else u''
        )
        for token_index in range(num_tokens)
    ]


def _get_tagged_tokens_or_empty(*args, **kwargs):
    try:
        return get_tagged_tokens(*args, **kwargs)
    except NoSuchTokenTagging:
        return []


def print_conll_style_tags_for_communication(
        comm, char_offsets=False, dependency=False, lemmas=False, ner=False,
        pos=False, other_tags=None,
        dependency_tool=None, lemmas_tool=None, ner_tool=None, pos_tool=None):

    """Print 'ConLL-style' tags for the tokens in a Communication

    Args:
        comm (Communication):
        char_offsets (bool): Flag for printing token text specified by
          a :class:`.Token`'s (optional) :class:`.TextSpan`
        dependency (bool): Flag for printing dependency parse HEAD tags
        lemmas (bool): Flag for printing lemma tags
        ner (bool): Flag for printing Named Entity Recognition tags
        pos (bool): Flag for printing Part-of-Speech tags
    """
    if other_tags is None:
        other_tags = ()

    header_fields = [u"INDEX", u"TOKEN"]
    if char_offsets:
        header_fields.append(u"CHAR")
    if lemmas:
        header_fields.append(u"LEMMA")
    if pos:
        header_fields.append(u"POS")
    if ner:
        header_fields.append(u"NER")
    if dependency:
        header_fields.append(u"HEAD")
        header_fields.append(u"DEPREL")
    for tag in other_tags:
        header_fields.append(tag)
    print(u"\t".join(header_fields))
    dashes = ["-" * len(fieldname) for fieldname in header_fields]
    print(u"\t".join(dashes))

    for tokenization in get_tokenizations(comm):
        token_tag_lists = []

        if char_offsets:
            token_tag_lists.append(
                _get_char_offset_tags_for_tokenization(comm, tokenization))
        if lemmas:
            token_tag_lists.append(
                _get_tagged_token_strs_by_token_index(
                    _get_tagged_tokens_or_empty(
                        tokenization, tagging_type=u'LEMMA', tool=lemmas_tool),
                    len(tokenization.tokenList.tokenList)))
        if pos:
            token_tag_lists.append(
                _get_tagged_token_strs_by_token_index(
                    _get_tagged_tokens_or_empty(
                        tokenization, tagging_type=u'POS', tool=pos_tool),
                    len(tokenization.tokenList.tokenList)))
        if ner:
            token_tag_lists.append([
                (tag if tag != u'NONE' else u'')
                for tag in _get_tagged_token_strs_by_token_index(
                    _get_tagged_tokens_or_empty(
                        tokenization, tagging_type=u'NER', tool=ner_tool),
                    len(tokenization.tokenList.tokenList)
                )
            ])
        if dependency:
            token_tag_lists.append(
                _get_conll_head_tags_for_tokenization(tokenization,
                                                      tool=dependency_tool))
            token_tag_lists.append(
                _get_conll_deprel_tags_for_tokenization(tokenization,
                                                        tool=dependency_tool))
        for tag in other_tags:
            token_tag_lists.append(
                _get_tagged_token_strs_by_token_index(
                    _get_tagged_tokens_or_empty(
                        tokenization, tagging_type=tag),
                    len(tokenization.tokenList.tokenList)))
        print_conll_style_tags_for_tokenization(tokenization,
                                                token_tag_lists)
        print()


def print_conll_style_tags_for_tokenization(tokenization, token_tag_lists):
    """Print 'ConLL-style' tags for the tokens in a tokenization

    Args:
        tokenization (Tokenization):
        token_tag_lists: A list of lists of token tag strings
    """
    if tokenization.tokenList:
        for i, token in enumerate(tokenization.tokenList.tokenList):
            token_tags = [unicode(token_tag_list[i])
                          for token_tag_list in token_tag_lists]
            fields = [unicode(i + 1), token.text]
            fields.extend(token_tags)
            print(u"\t".join(fields))


def _print_entity_mention_content(em, prefix=''):
    print(prefix + u"tokens:     %s" % (
        u" ".join(_get_tokens_for_entityMention(em))))
    if em.text:
        print(prefix + u"text:       %s" % em.text)
    print(prefix + u"entityType: %s" % em.entityType)
    print(prefix + u"phraseType: %s" % em.phraseType)


def print_entities(comm, tool=None):
    """Print information for :class:`.Entity` objects and their
    associated :class:`.EntityMention` objects

    Args:
        comm (Communication):
        tool (str): If not `None`, only print information for
                    :class:`.EntitySet` objects with a matching
                    `metadata.tool` field
    """
    if comm.entitySetList:
        for entitySet_index, entitySet in enumerate(comm.entitySetList):
            if tool is None or entitySet.metadata.tool == tool:
                print(u"Entity Set %d (%s):" % (entitySet_index,
                                                entitySet.metadata.tool))
                for entity_index, entity in enumerate(entitySet.entityList):
                    print(u"  Entity %d-%d:" % (entitySet_index, entity_index))
                    for em_index, em in enumerate(entity.mentionList):
                        print(u"      EntityMention %d-%d-%d:" % (
                            entitySet_index, entity_index, em_index))
                        _print_entity_mention_content(em, prefix=' ' * 10)
                        for (cm_index, cm) in enumerate(em.childMentionList):
                            print(u"          child EntityMention #%d:" % cm_index)
                            _print_entity_mention_content(cm, prefix=' ' * 14)
                    print()
                print()


def print_metadata(comm, tool=None):
    """Print metadata for tools used to annotate Communication

    Args:
        comm (Communication):
        tool (str): If not `None`, only print :class:`.AnnotationMetadata`
                    information for objects with a matching
                    `metadata.tool` field
    """
    if tool is None or comm.metadata.tool == tool:
        print(u"Communication:  %s\n" % comm.metadata.tool)

    dependency_parse_tools = set()
    parse_tools = set()
    tokenization_tools = set()
    token_tagging_tools = set()
    for tokenization in get_tokenizations(comm):
        tokenization_tools.add(tokenization.metadata.tool)
        if tokenization.tokenTaggingList:
            for tokenTagging in tokenization.tokenTaggingList:
                token_tagging_tools.add(tokenTagging.metadata.tool)
        if tokenization.dependencyParseList:
            for dependencyParse in tokenization.dependencyParseList:
                dependency_parse_tools.add(dependencyParse.metadata.tool)
        if tokenization.parseList:
            for parse in tokenization.parseList:
                parse_tools.add(parse.metadata.tool)

    communication_tagging_tools = set()
    for communication_tagging in lun(comm.communicationTaggingList):
        communication_tagging_tools.add(communication_tagging.metadata.tool)

    if tool is not None:
        dependency_parse_tools = dependency_parse_tools.intersection([tool])
        parse_tools = parse_tools.intersection([tool])
        tokenization_tools = tokenization_tools.intersection([tool])
        token_tagging_tools = token_tagging_tools.intersection([tool])
        communication_tagging_tools = communication_tagging_tools.intersection(
            [tool])

    if tokenization_tools:
        for toolname in sorted(tokenization_tools):
            print(u"  Tokenization:  %s" % toolname)
        print()
    if dependency_parse_tools:
        for toolname in sorted(dependency_parse_tools):
            print(u"    Dependency Parse:  %s" % toolname)
        print()
    if parse_tools:
        for toolname in sorted(parse_tools):
            print(u"    Parse:  %s" % toolname)
        print()
    if token_tagging_tools:
        for toolname in sorted(token_tagging_tools):
            print(u"    TokenTagging:  %s" % toolname)
        print()

    if comm.entityMentionSetList:
        for i, em_set in enumerate(comm.entityMentionSetList):
            if tool is None or em_set.metadata.tool == tool:
                print(u"  EntityMentionSet #%d:  %s" % (
                    i, em_set.metadata.tool))
        print()
    if comm.entitySetList:
        for i, entitySet in enumerate(comm.entitySetList):
            if tool is None or entitySet.metadata.tool == tool:
                print(u"  EntitySet #%d:  %s" % (
                    i, entitySet.metadata.tool))
        print()
    if comm.situationMentionSetList:
        for i, sm_set in enumerate(comm.situationMentionSetList):
            if tool is None or sm_set.metadata.tool == tool:
                print(u"  SituationMentionSet #%d:  %s" % (
                    i, sm_set.metadata.tool))
        print()
    if comm.situationSetList:
        for i, situationSet in enumerate(comm.situationSetList):
            if tool is None or situationSet.metadata.tool == tool:
                print(u"  SituationSet #%d:  %s" % (
                    i, situationSet.metadata.tool))
        print()

    if communication_tagging_tools:
        for toolname in sorted(communication_tagging_tools):
            print(u"  CommunicationTagging:  %s" % toolname)
        print()


def print_sections(comm, tool=None):
    """Print information for all :class:`.Section` object, according to their spans.

    Args:
        comm (Communication):
        tool (str): If not `None`, only print information for
                    :class:`.Section` objects with a matching
                    `metadata.tool` field
    """
    if tool is None or comm.metadata.tool == tool:
        text = comm.text
        for sect_idx, sect in enumerate(lun(comm.sectionList)):
            ts = sect.textSpan
            if ts is None:
                print(u"Section %s does not have a textSpan ")
                "field set" % (sect.uuid.uuidString)
                continue
            print(u"Section %d (%s), from %d to %d:" % (
                sect_idx, sect.uuid.uuidString, ts.start, ts.ending))
            print(u"%s" % (text[ts.start:ts.ending]))
            print()
        print()


def print_situation_mentions(comm, tool=None):
    """Print information for all :class:`.SituationMention` (some of which
    may not have a :class:`.Situation`)

    Args:
        comm (Communication):
        tool (str): If not `None`, only print information for
                    :class:`.SituationMention` objects with a matching
                    `metadata.tool` field
    """
    for sm_set_idx, sm_set in enumerate(lun(comm.situationMentionSetList)):
        if tool is None or sm_set.metadata.tool == tool:
            print(u"Situation Set %d (%s):" % (sm_set_idx,
                                               sm_set.metadata.tool))
            for sm_idx, sm in enumerate(sm_set.mentionList):
                print(u"  SituationMention %d-%d:" % (sm_set_idx, sm_idx))
                _print_situation_mention(sm)
                print()
            print()


def print_situations(comm, tool=None):
    """Print information for all :class:`.Situation` objects and their
    associated :class:`.SituationMention` objects

    Args:
        comm (Communication):
        tool (str): If not `None`, only print information for
                    :class:`.Situation` objects with a matching
                    `metadata.tool` field
    """
    for s_set_idx, s_set in enumerate(lun(comm.situationSetList)):
        if tool is None or s_set.metadata.tool == tool:
            print(u"Situation Set %d (%s):" % (s_set_idx,
                                               s_set.metadata.tool))
            for s_idx, situation in enumerate(s_set.situationList):
                print(u"  Situation %d-%d:" % (s_set_idx, s_idx))
                _p(6, 18, u"situationType", situation.situationType)
                for sm_idx, sm in enumerate(lun(situation.mentionList)):
                    print(u" " * 6 + u"SituationMention %d-%d-%d:" % (
                        s_set_idx, s_idx, sm_idx))
                    _print_situation_mention(sm)
                print()
            print()


def _print_situation_mention(situationMention):
    """Helper function for printing info for a SituationMention"""
    if situationMention.text:
        _p(10, 20, u"text", situationMention.text)
    if situationMention.situationType:
        _p(10, 20, u"situationType", situationMention.situationType)
    for arg_idx, ma in enumerate(lun(situationMention.argumentList)):
        print(u" " * 10 + u"Argument %d:" % arg_idx)
        if ma.role:
            _p(14, 16, u"role", ma.role)
        if ma.entityMention:
            _p(14, 16, u"entityMention",
                u" ".join(_get_tokens_for_entityMention(ma.entityMention)))
        if ma.propertyList:
            # PROTO-ROLE PROPERTIES: Format a separate list for each
            # distinct annotator (metadata.tool) which tool should be
            # either None or a string. Sort by annotator
            # (metadata.tool) and then by property (p.value)
            last_tool = False
            for p in sorted(ma.propertyList,
                            key=lambda x: (x.metadata.tool, x.value)):
                tool = p.metadata.tool
                if tool != last_tool:
                    print(u" " * 14 + u"Properties (%s):" % tool)
                    last_tool = tool
                _p(18, 20, p.value, u"%1.1f" % p.polarity)
        # A SituationMention can have an argumentList with a
        # MentionArgument that points to another SituationMention---
        # which could conceivably lead to loops.  We currently don't
        # traverse the list recursively, instead looking at only
        # SituationMentions referenced by top-level SituationMentions
        if ma.situationMention:
            print(u" " * 14 + u"situationMention:")
            if situationMention.text:
                _p(18, 20, u"text", situationMention.text)
            if situationMention.situationType:
                _p(18, 20, u"situationType", situationMention.situationType)


def _p(indent_level, justified_width, fieldname, content):
    """Text alignment helper function"""
    print (
        (u" " * indent_level) +
        (fieldname + u":").ljust(justified_width) +
        content
    )


def print_text_for_communication(comm, tool=None):
    """Print `text field of :class:`.Communication`

    Args:
        comm (Communication):
        tool (str): If not `None`, only print `text` field of
                    :class:`.Communication` objects with a matching
                    `metadata.tool` field
    """
    if tool is None or comm.metadata.tool == tool:
        print(comm.text)


def print_id_for_communication(comm, tool=None):
    """Print ID field of :class:`.Communication`

    Args:
        comm (Communication):
        tool (str): If not `None`, only print ID of
                    :class:`.Communication` objects with a matching
                    `metadata.tool` field

    """
    if tool is None or comm.metadata.tool == tool:
        print(comm.id)


def print_communication_taggings_for_communication(comm, tool=None):
    """Print information for :class:`.CommunicationTagging` objects

    Args:
        comm (Communication):
        tool (str): If not `None`, only print information for
                    :class:`.CommunicationTagging` objects with a
                    matching `metadata.tool` field
    """
    communication_taggings = _filter_by_tool(
        lun(comm.communicationTaggingList), tool)
    for tagging in communication_taggings:
        print('%s: %s' % (
            tagging.taggingType,
            ' '.join('%s:%.3f' % p for p in
                     zip(tagging.tagList, tagging.confidenceList))
        ))


def print_tokens_with_entityMentions(comm, tool=None):
    """Print information for :class:`.Token` objects that are part of an :class:`.EntityMention`

    Args:
        comm (Communication):
        tool (str): If not `None`, only print information for tokens
                    that are associated with an
                    :class:`.EntityMention` that is part of an
                    :class:`.EntityMentionSet` with a matching
                    `metadata.tool` field
    """
    em_by_tkzn_id = _get_entityMentions_by_tokenizationId(
        comm, tool=tool)
    em_entity_num = _get_entity_number_for_entityMention_uuid(comm, tool=tool)
    tokenizations_by_section = _get_tokenizations_grouped_by_section(comm)

    for tokenizations_in_section in tokenizations_by_section:
        for tokenization in tokenizations_in_section:
            if tokenization.tokenList:
                text_tokens = [token.text
                               for token
                               in tokenization.tokenList.tokenList]
                u = tokenization.uuid.uuidString
                if u in em_by_tkzn_id:
                    for em in em_by_tkzn_id[u]:
                        first_token_index = em.tokens.tokenIndexList[0]
                        last_token_index = em.tokens.tokenIndexList[-1]
                        entity_number = em_entity_num[em.uuid.uuidString]
                        text_tokens[first_token_index] = (
                            u"<ENTITY ID=%d>%s" %
                            (entity_number, text_tokens[first_token_index])
                        )
                        text_tokens[last_token_index] = (
                            u"%s</ENTITY>" % text_tokens[last_token_index]
                        )
                print(u" ".join(text_tokens))
        print()


def print_tokens_for_communication(comm, tool=None):
    """Print token text for a :class:`.Communication`

    Args:
        comm (Communication):
        tool (str): If not `None`, only print token text for
                    :class:`.Communication` objects with a matching
                    `metadata.tool` field
    """
    tokenizations_by_section = _get_tokenizations_grouped_by_section(
        comm, tool=tool)

    for tokenizations_in_section in tokenizations_by_section:
        for tokenization in tokenizations_in_section:
            if tokenization.tokenList:
                text_tokens = [token.text
                               for token
                               in tokenization.tokenList.tokenList]
                print(u" ".join(text_tokens))
        print()


def print_penn_treebank_for_communication(comm, tool=None):
    """Print Penn-Treebank parse trees for all :class:`.Tokenization` objects

    Args:
        comm (Communication):
        tool (str): If not `None`, only print information for
                    :class:`.Tokenization` objects with a matching
                    `metadata.tool` field
    """
    tokenizations = get_tokenizations(comm)

    for tokenization in tokenizations:
        if tokenization.parseList:
            for parse in tokenization.parseList:
                if tool is None or tool == parse.metadata.tool:
                    print(penn_treebank_for_parse(parse) + u"\n\n")


def penn_treebank_for_parse(parse):
    """Get a Penn-Treebank style string for a Concrete Parse object

    Args:
        parse (Parse):

    Returns:
        str: A string containing a Penn Treebank style parse tree representation
    """
    def _traverse_parse(nodes, node_index, indent=0):
        s = u""
        indent += len(nodes[node_index].tag) + 2
        if nodes[node_index].childList:
            s += u"(%s " % nodes[node_index].tag
            for i, child_node_index in enumerate(nodes[node_index].childList):
                if i > 0:
                    s += u"\n" + u" " * indent
                s += _traverse_parse(nodes, child_node_index, indent)
            s += u")"
        else:
            s += nodes[node_index].tag
        return s

    sorted_nodes = sorted(parse.constituentList, key=attrgetter('id'))
    return _traverse_parse(sorted_nodes, 0)


def _get_char_offset_tags_for_tokenization(comm, tokenization):
    if tokenization.tokenList:
        char_offset_tags = [None] * len(tokenization.tokenList.tokenList)

        if comm.text:
            for i, token in enumerate(tokenization.tokenList.tokenList):
                if token.textSpan:
                    char_offset_tags[i] = comm.text[
                        token.textSpan.start:token.textSpan.ending]
        return char_offset_tags


def _deps_for_tokenization(tokenization,
                           dependency_parse_index=0,
                           tool=None):
    """
    Return a generator of the dependencies (Dependency objects) for
    a tokenization under the given tool.
    """
    if tokenization.tokenList is not None:
        # Tokens that are not part of the dependency parse
        # (e.g. punctuation) are represented using an empty string
        dp_idx = _reconcile_index_and_tool(tokenization.dependencyParseList,
                                           dependency_parse_index,
                                           tool)

        if _valid_index_lun(tokenization.dependencyParseList, dp_idx):
            dp = tokenization.dependencyParseList[dp_idx]
            if tool is None or dp.metadata.tool == tool:
                for dependency in dp.dependencyList:
                    yield dependency


def _sorted_dep_list_for_tokenization(tokenization,
                                      dependency_parse_index=0,
                                      tool=None):
    """
    Return output of _deps_for_tokenization in a list whose length
    is equal to the number of tokens in this tokenization's token list,
    where the element at index i is a dependency if there is a
    dependency whose dep field is i and None otherwise.
    """
    if tokenization.tokenList is not None:
        dep_list = [None] * len(tokenization.tokenList.tokenList)
        for dep in _deps_for_tokenization(
                tokenization, dependency_parse_index=dependency_parse_index,
                tool=tool):
            dep_list[dep.dep] = dep
        return dep_list
    else:
        return []


def _get_conll_head_tags_for_tokenization(tokenization,
                                          dependency_parse_index=0,
                                          tool=None):
    """Get a list of ConLL 'HEAD tags' for a tokenization

    In the ConLL data format:

        http://ufal.mff.cuni.cz/conll2009-st/task-description.html

    the HEAD for a token is the (1-indexed) index of that token's
    parent token.  The root token of the dependency parse has a HEAD
    index of 0.

    Args:
        tokenization (Tokenization):

    Returns:
        str[]: A list of ConLL 'HEAD tag' strings, with one HEAD tag
               for each token in the supplied tokenization.  If a
               token does not have a HEAD tag (e.g. punctuation
               tokens), the HEAD tag is an empty string.

               If the tokenization does not have a Dependency Parse,
               this function returns a list of empty strings for each
               token in the supplied tokenization.
    """
    return list(map(
        lambda dep: '' if dep is None else (
            0 if dep.gov is None else dep.gov + 1),
        _sorted_dep_list_for_tokenization(
            tokenization, dependency_parse_index=dependency_parse_index,
            tool=tool)))


def _get_conll_deprel_tags_for_tokenization(tokenization,
                                            dependency_parse_index=0,
                                            tool=None):
    """Get a list of ConLL 'DEPREL tags' for a tokenization

    In the ConLL data format:

        http://ufal.mff.cuni.cz/conll2009-st/task-description.html

    the DEPREL for a token is the type of that token's dependency with
    its parent.

    Args:
        tokenization (Tokenization):

    Returns:
        A list of ConLL 'DEPREL tag' strings, with one DEPREL tag for
        each token in the supplied tokenization.  If a token does not
        have a DEPREL tag (e.g. punctuation tokens), the DEPREL tag is
        an empty string.

        If the tokenization does not have a Dependency Parse, this
        function returns a list of empty strings for each token in the
        supplied tokenization.
    """
    return list(map(
        lambda dep: '' if dep is None else (
            '' if dep.edgeType is None else dep.edgeType),
        _sorted_dep_list_for_tokenization(
            tokenization, dependency_parse_index=dependency_parse_index,
            tool=tool)))


def _get_entityMentions_by_tokenizationId(comm, tool=None):
    """Get entity mentions for a Communication grouped by Tokenization
    UUID string

    Args:
        comm (Communication):

    Returns:
        A dictionary of lists of EntityMentions, where the dictionary
        keys are Tokenization UUID strings.
    """
    mentions_by_tkzn_id = defaultdict(list)
    for entitySet in lun(comm.entitySetList):
        for entity in entitySet.entityList:
            for entityMention in entity.mentionList:
                if (tool is None or
                        entityMention.entityMentionSet.metadata.tool == tool):
                    u = entityMention.tokens.tokenizationId.uuidString
                    mentions_by_tkzn_id[u].append(entityMention)
    return mentions_by_tkzn_id


def _get_entity_number_for_entityMention_uuid(comm, tool=None):
    """Create mapping from EntityMention UUID to (zero-indexed)
    'Entity Number'

    Args:
        comm (Communication):

    Returns:
        A dictionary where the keys are EntityMention UUID strings,
        and the values are "Entity Numbers", where the first Entity is
        assigned number 0, the second Entity is assigned number 1,
        etc.
    """
    entity_number_for_entityMention_uuid = {}
    entity_number_counter = 0

    if comm.entitySetList:
        for entitySet in comm.entitySetList:
            for entity in entitySet.entityList:
                any_mention = False
                for entityMention in entity.mentionList:
                    if (tool is None or
                            entityMention.entityMentionSet.metadata.tool ==
                            tool):
                        entity_number_for_entityMention_uuid[
                            entityMention.uuid.uuidString
                        ] = entity_number_counter
                    any_mention = True
                if any_mention:
                    entity_number_counter += 1
    return entity_number_for_entityMention_uuid


def _get_tokenizations_grouped_by_section(comm, tool=None):
    """Returns a list of lists of Tokenization objects in a Communication

    Args:
        comm (Communication):

    Returns:
        Returns a list of lists of Tokenization objects, where the
        Tokenization objects are grouped by Section
    """
    tokenizations_by_section = []

    if comm.sectionList:
        for section in comm.sectionList:
            tokenizations_in_section = []
            if section.sentenceList:
                for sentence in section.sentenceList:
                    if sentence.tokenization:
                        if (tool is None or
                                sentence.tokenization.metadata.tool == tool):
                            tokenizations_in_section.append(
                                sentence.tokenization)
            tokenizations_by_section.append(tokenizations_in_section)

    return tokenizations_by_section


def _get_tokens_for_entityMention(entityMention):
    """Get list of token strings for an EntityMention

    Args:
        entityMention (EntityMention):

    Returns:
        A list of token strings
    """
    tokens = []
    for tokenIndex in entityMention.tokens.tokenIndexList:
        tokens.append(entityMention.tokens.tokenization.tokenList.tokenList[
                      tokenIndex].text)
    return tokens


def _get_tokentaggings_of_type(tokenization, taggingType, tool=None):
    """Returns a list of :class:`.TokenTagging` objects with the specified taggingType

    Args:
        tokenization (Tokenization):
        taggingType (str): A string value for the specified TokenTagging.taggingType

    Returns:
        A list of TokenTagging objects
    """
    return [
        tt for tt in tokenization.tokenTaggingList
        if tt.taggingType.lower() == taggingType.lower() and (
            tool is None or tt.metadata.tool == tool)
    ]



# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.10.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
import sys
import concrete.structure.ttypes
import concrete.metadata.ttypes
import concrete.uuid.ttypes
import concrete.linking.ttypes

from thrift.transport import TTransport


class Property(object):
    """
    Attached to Arguments to support situations where
    a 'participant' has more than one 'property' (in BinarySRL terms),
    whereas Arguments notionally only support one Role.

    Attributes:
     - value: The required value of the property.
     - metadata: Metadata to support this particular property object.
     - polarity: This value is typically boolean, 0.0 or 1.0, but we use a
    float in order to potentially capture cases where an annotator is
    highly confident that the value is underspecified, via a value of
    0.5.
    """

    thrift_spec = (
        None,  # 0
        (1, TType.STRING, 'value', 'UTF8', None, ),  # 1
        (2, TType.STRUCT, 'metadata', (concrete.metadata.ttypes.AnnotationMetadata, concrete.metadata.ttypes.AnnotationMetadata.thrift_spec), None, ),  # 2
        (3, TType.DOUBLE, 'polarity', None, None, ),  # 3
    )

    def __init__(self, value=None, metadata=None, polarity=None,):
        self.value = value
        self.metadata = metadata
        self.polarity = polarity

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.value = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRUCT:
                    self.metadata = concrete.metadata.ttypes.AnnotationMetadata()
                    self.metadata.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.DOUBLE:
                    self.polarity = iprot.readDouble()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('Property')
        if self.value is not None:
            oprot.writeFieldBegin('value', TType.STRING, 1)
            oprot.writeString(self.value.encode('utf-8') if sys.version_info[0] == 2 else self.value)
            oprot.writeFieldEnd()
        if self.metadata is not None:
            oprot.writeFieldBegin('metadata', TType.STRUCT, 2)
            self.metadata.write(oprot)
            oprot.writeFieldEnd()
        if self.polarity is not None:
            oprot.writeFieldBegin('polarity', TType.DOUBLE, 3)
            oprot.writeDouble(self.polarity)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.value is None:
            raise TProtocolException(message='Required field value is unset!')
        if self.metadata is None:
            raise TProtocolException(message='Required field metadata is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Argument(object):
    """
    A situation argument, consisting of an argument role and a value.
    Argument values may be Entities or Situations.

    Attributes:
     - role: The relationship between this argument and the situation that
    owns it. The roles that a situation's arguments can take
    depend on the type of the situation (including subtype
    information, such as event_type).
     - entityId: A pointer to the value of this argument, if it is explicitly
    encoded as an Entity.
     - situationId: A pointer to the value of this argument, if it is a situation.
     - propertyList: For the BinarySRL task, there may be situations
    where more than one property is attached to a single
    participant. A list of these properties can be stored in this field.
    """

    thrift_spec = (
        None,  # 0
        (1, TType.STRING, 'role', 'UTF8', None, ),  # 1
        (2, TType.STRUCT, 'entityId', (concrete.uuid.ttypes.UUID, concrete.uuid.ttypes.UUID.thrift_spec), None, ),  # 2
        (3, TType.STRUCT, 'situationId', (concrete.uuid.ttypes.UUID, concrete.uuid.ttypes.UUID.thrift_spec), None, ),  # 3
        (4, TType.LIST, 'propertyList', (TType.STRUCT, (Property, Property.thrift_spec), False), None, ),  # 4
    )

    def __init__(self, role=None, entityId=None, situationId=None, propertyList=None,):
        self.role = role
        self.entityId = entityId
        self.situationId = situationId
        self.propertyList = propertyList

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.role = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRUCT:
                    self.entityId = concrete.uuid.ttypes.UUID()
                    self.entityId.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRUCT:
                    self.situationId = concrete.uuid.ttypes.UUID()
                    self.situationId.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.LIST:
                    self.propertyList = []
                    (_etype3, _size0) = iprot.readListBegin()
                    for _i4 in range(_size0):
                        _elem5 = Property()
                        _elem5.read(iprot)
                        self.propertyList.append(_elem5)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('Argument')
        if self.role is not None:
            oprot.writeFieldBegin('role', TType.STRING, 1)
            oprot.writeString(self.role.encode('utf-8') if sys.version_info[0] == 2 else self.role)
            oprot.writeFieldEnd()
        if self.entityId is not None:
            oprot.writeFieldBegin('entityId', TType.STRUCT, 2)
            self.entityId.write(oprot)
            oprot.writeFieldEnd()
        if self.situationId is not None:
            oprot.writeFieldBegin('situationId', TType.STRUCT, 3)
            self.situationId.write(oprot)
            oprot.writeFieldEnd()
        if self.propertyList is not None:
            oprot.writeFieldBegin('propertyList', TType.LIST, 4)
            oprot.writeListBegin(TType.STRUCT, len(self.propertyList))
            for iter6 in self.propertyList:
                iter6.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Justification(object):
    """
    Attributes:
     - justificationType: An enumerated value used to describe the way in which the
    justification's mention provides supporting evidence for the
    situation.
     - mentionId: A pointer to the SituationMention itself.
     - tokenRefSeqList: An optional list of pointers to tokens that are (especially)
    relevant to the way in which this mention provides
    justification for the situation. It is left up to individual
    analytics to decide what tokens (if any) they wish to include
    in this field.
    """

    thrift_spec = (
        None,  # 0
        (1, TType.STRING, 'justificationType', 'UTF8', None, ),  # 1
        (2, TType.STRUCT, 'mentionId', (concrete.uuid.ttypes.UUID, concrete.uuid.ttypes.UUID.thrift_spec), None, ),  # 2
        (3, TType.LIST, 'tokenRefSeqList', (TType.STRUCT, (concrete.structure.ttypes.TokenRefSequence, concrete.structure.ttypes.TokenRefSequence.thrift_spec), False), None, ),  # 3
    )

    def __init__(self, justificationType=None, mentionId=None, tokenRefSeqList=None,):
        self.justificationType = justificationType
        self.mentionId = mentionId
        self.tokenRefSeqList = tokenRefSeqList

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.justificationType = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRUCT:
                    self.mentionId = concrete.uuid.ttypes.UUID()
                    self.mentionId.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.LIST:
                    self.tokenRefSeqList = []
                    (_etype10, _size7) = iprot.readListBegin()
                    for _i11 in range(_size7):
                        _elem12 = concrete.structure.ttypes.TokenRefSequence()
                        _elem12.read(iprot)
                        self.tokenRefSeqList.append(_elem12)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('Justification')
        if self.justificationType is not None:
            oprot.writeFieldBegin('justificationType', TType.STRING, 1)
            oprot.writeString(self.justificationType.encode('utf-8') if sys.version_info[0] == 2 else self.justificationType)
            oprot.writeFieldEnd()
        if self.mentionId is not None:
            oprot.writeFieldBegin('mentionId', TType.STRUCT, 2)
            self.mentionId.write(oprot)
            oprot.writeFieldEnd()
        if self.tokenRefSeqList is not None:
            oprot.writeFieldBegin('tokenRefSeqList', TType.LIST, 3)
            oprot.writeListBegin(TType.STRUCT, len(self.tokenRefSeqList))
            for iter13 in self.tokenRefSeqList:
                iter13.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.mentionId is None:
            raise TProtocolException(message='Required field mentionId is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class TimeML(object):
    """
    A wrapper for various TimeML annotations.

    Attributes:
     - timeMLClass: The TimeML class for situations representing TimeML events
     - timeMLTense: The TimeML tense for situations representing TimeML events
     - timeMLAspect: The TimeML aspect for situations representing TimeML events
    """

    thrift_spec = (
        None,  # 0
        (1, TType.STRING, 'timeMLClass', 'UTF8', None, ),  # 1
        (2, TType.STRING, 'timeMLTense', 'UTF8', None, ),  # 2
        (3, TType.STRING, 'timeMLAspect', 'UTF8', None, ),  # 3
    )

    def __init__(self, timeMLClass=None, timeMLTense=None, timeMLAspect=None,):
        self.timeMLClass = timeMLClass
        self.timeMLTense = timeMLTense
        self.timeMLAspect = timeMLAspect

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.timeMLClass = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.timeMLTense = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.timeMLAspect = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('TimeML')
        if self.timeMLClass is not None:
            oprot.writeFieldBegin('timeMLClass', TType.STRING, 1)
            oprot.writeString(self.timeMLClass.encode('utf-8') if sys.version_info[0] == 2 else self.timeMLClass)
            oprot.writeFieldEnd()
        if self.timeMLTense is not None:
            oprot.writeFieldBegin('timeMLTense', TType.STRING, 2)
            oprot.writeString(self.timeMLTense.encode('utf-8') if sys.version_info[0] == 2 else self.timeMLTense)
            oprot.writeFieldEnd()
        if self.timeMLAspect is not None:
            oprot.writeFieldBegin('timeMLAspect', TType.STRING, 3)
            oprot.writeString(self.timeMLAspect.encode('utf-8') if sys.version_info[0] == 2 else self.timeMLAspect)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Situation(object):
    """
    A single situation, along with pointers to situation mentions that
    provide evidence for the situation. "Situations" include events,
    relations, facts, sentiments, and beliefs. Each situation has a
    core type (such as EVENT or SENTIMENT), along with an optional
    subtype based on its core type (e.g., event_type=CONTACT_MEET), and
    a set of zero or more unordered arguments.

    This struct may be used for a variety of "processed" Situations such
    as (but not limited to):
    - SituationMentions which have been collapsed into a coreferential cluster
    - Situations which are inferred and not directly supported by a textual mention

    Attributes:
     - uuid: Unique identifier for this situation.
     - situationType: The core type of this situation (eg EVENT or SENTIMENT),
    or a coarse grain situation type.
     - situationKind: A fine grain situation type that specifically describes the
    situation based on situationType above. It allows for more
    detailed description of the situation.

    Some examples:

    if situationType == EVENT, the event type for the situation
    if situationType == STATE, the state type
    if situationType == TEMPORAL_FACT, the temporal fact type

    For Propbank, this field should be the predicate lemma and id,
    e.g. "strike.02". For FrameNet, this should be the frame name,
    e.g. "Commerce_buy".

    Different and more varied situationTypes may be added
    in the future.
     - argumentList: The arguments for this situation. Each argument consists of a
    role and a value. It is possible for an situation to have
    multiple arguments with the same role. Arguments are
    unordered.
     - mentionIdList: Ids of the mentions of this situation in a communication
    (type=SituationMention)
     - justificationList: An list of pointers to SituationMentions that provide
    justification for this situation. These mentions may be either
    direct mentions of the situation, or indirect evidence.
     - timeML: A wrapper for TimeML annotations.
     - intensity: An "intensity" rating for this situation, typically ranging from
    0-1. In the case of SENTIMENT situations, this is used to record
    the intensity of the sentiment.
     - polarity: The polarity of this situation. In the case of SENTIMENT
    situations, this is used to record the polarity of the
    sentiment.
     - confidence: A confidence score for this individual situation. You can also
    set a confidence score for an entire SituationSet using the
    SituationSet's metadata.
    """

    thrift_spec = (
        None,  # 0
        (1, TType.STRUCT, 'uuid', (concrete.uuid.ttypes.UUID, concrete.uuid.ttypes.UUID.thrift_spec), None, ),  # 1
        (2, TType.STRING, 'situationType', 'UTF8', None, ),  # 2
        (3, TType.LIST, 'argumentList', (TType.STRUCT, (Argument, Argument.thrift_spec), False), None, ),  # 3
        (4, TType.LIST, 'mentionIdList', (TType.STRUCT, (concrete.uuid.ttypes.UUID, concrete.uuid.ttypes.UUID.thrift_spec), False), None, ),  # 4
        (5, TType.LIST, 'justificationList', (TType.STRUCT, (Justification, Justification.thrift_spec), False), None, ),  # 5
        None,  # 6
        None,  # 7
        None,  # 8
        None,  # 9
        None,  # 10
        None,  # 11
        None,  # 12
        None,  # 13
        None,  # 14
        None,  # 15
        None,  # 16
        None,  # 17
        None,  # 18
        None,  # 19
        None,  # 20
        None,  # 21
        None,  # 22
        None,  # 23
        None,  # 24
        None,  # 25
        None,  # 26
        None,  # 27
        None,  # 28
        None,  # 29
        None,  # 30
        None,  # 31
        None,  # 32
        None,  # 33
        None,  # 34
        None,  # 35
        None,  # 36
        None,  # 37
        None,  # 38
        None,  # 39
        None,  # 40
        None,  # 41
        None,  # 42
        None,  # 43
        None,  # 44
        None,  # 45
        None,  # 46
        None,  # 47
        None,  # 48
        None,  # 49
        (50, TType.STRING, 'situationKind', 'UTF8', None, ),  # 50
        None,  # 51
        None,  # 52
        None,  # 53
        (54, TType.STRUCT, 'timeML', (TimeML, TimeML.thrift_spec), None, ),  # 54
        None,  # 55
        None,  # 56
        None,  # 57
        None,  # 58
        None,  # 59
        None,  # 60
        None,  # 61
        None,  # 62
        None,  # 63
        None,  # 64
        None,  # 65
        None,  # 66
        None,  # 67
        None,  # 68
        None,  # 69
        None,  # 70
        None,  # 71
        None,  # 72
        None,  # 73
        None,  # 74
        None,  # 75
        None,  # 76
        None,  # 77
        None,  # 78
        None,  # 79
        None,  # 80
        None,  # 81
        None,  # 82
        None,  # 83
        None,  # 84
        None,  # 85
        None,  # 86
        None,  # 87
        None,  # 88
        None,  # 89
        None,  # 90
        None,  # 91
        None,  # 92
        None,  # 93
        None,  # 94
        None,  # 95
        None,  # 96
        None,  # 97
        None,  # 98
        None,  # 99
        (100, TType.DOUBLE, 'intensity', None, None, ),  # 100
        (101, TType.STRING, 'polarity', 'UTF8', None, ),  # 101
        None,  # 102
        None,  # 103
        None,  # 104
        None,  # 105
        None,  # 106
        None,  # 107
        None,  # 108
        None,  # 109
        None,  # 110
        None,  # 111
        None,  # 112
        None,  # 113
        None,  # 114
        None,  # 115
        None,  # 116
        None,  # 117
        None,  # 118
        None,  # 119
        None,  # 120
        None,  # 121
        None,  # 122
        None,  # 123
        None,  # 124
        None,  # 125
        None,  # 126
        None,  # 127
        None,  # 128
        None,  # 129
        None,  # 130
        None,  # 131
        None,  # 132
        None,  # 133
        None,  # 134
        None,  # 135
        None,  # 136
        None,  # 137
        None,  # 138
        None,  # 139
        None,  # 140
        None,  # 141
        None,  # 142
        None,  # 143
        None,  # 144
        None,  # 145
        None,  # 146
        None,  # 147
        None,  # 148
        None,  # 149
        None,  # 150
        None,  # 151
        None,  # 152
        None,  # 153
        None,  # 154
        None,  # 155
        None,  # 156
        None,  # 157
        None,  # 158
        None,  # 159
        None,  # 160
        None,  # 161
        None,  # 162
        None,  # 163
        None,  # 164
        None,  # 165
        None,  # 166
        None,  # 167
        None,  # 168
        None,  # 169
        None,  # 170
        None,  # 171
        None,  # 172
        None,  # 173
        None,  # 174
        None,  # 175
        None,  # 176
        None,  # 177
        None,  # 178
        None,  # 179
        None,  # 180
        None,  # 181
        None,  # 182
        None,  # 183
        None,  # 184
        None,  # 185
        None,  # 186
        None,  # 187
        None,  # 188
        None,  # 189
        None,  # 190
        None,  # 191
        None,  # 192
        None,  # 193
        None,  # 194
        None,  # 195
        None,  # 196
        None,  # 197
        None,  # 198
        None,  # 199
        (200, TType.DOUBLE, 'confidence', None, None, ),  # 200
    )

    def __init__(self, uuid=None, situationType=None, situationKind=None, argumentList=None, mentionIdList=None, justificationList=None, timeML=None, intensity=None, polarity=None, confidence=None,):
        self.uuid = uuid
        self.situationType = situationType
        self.situationKind = situationKind
        self.argumentList = argumentList
        self.mentionIdList = mentionIdList
        self.justificationList = justificationList
        self.timeML = timeML
        self.intensity = intensity
        self.polarity = polarity
        self.confidence = confidence

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.uuid = concrete.uuid.ttypes.UUID()
                    self.uuid.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.situationType = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 50:
                if ftype == TType.STRING:
                    self.situationKind = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.LIST:
                    self.argumentList = []
                    (_etype17, _size14) = iprot.readListBegin()
                    for _i18 in range(_size14):
                        _elem19 = Argument()
                        _elem19.read(iprot)
                        self.argumentList.append(_elem19)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.LIST:
                    self.mentionIdList = []
                    (_etype23, _size20) = iprot.readListBegin()
                    for _i24 in range(_size20):
                        _elem25 = concrete.uuid.ttypes.UUID()
                        _elem25.read(iprot)
                        self.mentionIdList.append(_elem25)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.LIST:
                    self.justificationList = []
                    (_etype29, _size26) = iprot.readListBegin()
                    for _i30 in range(_size26):
                        _elem31 = Justification()
                        _elem31.read(iprot)
                        self.justificationList.append(_elem31)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 54:
                if ftype == TType.STRUCT:
                    self.timeML = TimeML()
                    self.timeML.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 100:
                if ftype == TType.DOUBLE:
                    self.intensity = iprot.readDouble()
                else:
                    iprot.skip(ftype)
            elif fid == 101:
                if ftype == TType.STRING:
                    self.polarity = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 200:
                if ftype == TType.DOUBLE:
                    self.confidence = iprot.readDouble()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('Situation')
        if self.uuid is not None:
            oprot.writeFieldBegin('uuid', TType.STRUCT, 1)
            self.uuid.write(oprot)
            oprot.writeFieldEnd()
        if self.situationType is not None:
            oprot.writeFieldBegin('situationType', TType.STRING, 2)
            oprot.writeString(self.situationType.encode('utf-8') if sys.version_info[0] == 2 else self.situationType)
            oprot.writeFieldEnd()
        if self.argumentList is not None:
            oprot.writeFieldBegin('argumentList', TType.LIST, 3)
            oprot.writeListBegin(TType.STRUCT, len(self.argumentList))
            for iter32 in self.argumentList:
                iter32.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.mentionIdList is not None:
            oprot.writeFieldBegin('mentionIdList', TType.LIST, 4)
            oprot.writeListBegin(TType.STRUCT, len(self.mentionIdList))
            for iter33 in self.mentionIdList:
                iter33.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.justificationList is not None:
            oprot.writeFieldBegin('justificationList', TType.LIST, 5)
            oprot.writeListBegin(TType.STRUCT, len(self.justificationList))
            for iter34 in self.justificationList:
                iter34.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.situationKind is not None:
            oprot.writeFieldBegin('situationKind', TType.STRING, 50)
            oprot.writeString(self.situationKind.encode('utf-8') if sys.version_info[0] == 2 else self.situationKind)
            oprot.writeFieldEnd()
        if self.timeML is not None:
            oprot.writeFieldBegin('timeML', TType.STRUCT, 54)
            self.timeML.write(oprot)
            oprot.writeFieldEnd()
        if self.intensity is not None:
            oprot.writeFieldBegin('intensity', TType.DOUBLE, 100)
            oprot.writeDouble(self.intensity)
            oprot.writeFieldEnd()
        if self.polarity is not None:
            oprot.writeFieldBegin('polarity', TType.STRING, 101)
            oprot.writeString(self.polarity.encode('utf-8') if sys.version_info[0] == 2 else self.polarity)
            oprot.writeFieldEnd()
        if self.confidence is not None:
            oprot.writeFieldBegin('confidence', TType.DOUBLE, 200)
            oprot.writeDouble(self.confidence)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.uuid is None:
            raise TProtocolException(message='Required field uuid is unset!')
        if self.situationType is None:
            raise TProtocolException(message='Required field situationType is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class SituationSet(object):
    """
    A theory about the set of situations that are present in a
    message. See also: Situation

    Attributes:
     - uuid: Unique identifier for this set.
     - metadata: Information about where this set came from.
     - situationList: List of mentions in this set.
     - linkingList: Entity linking annotations associated with this SituationSet.
    """

    thrift_spec = (
        None,  # 0
        (1, TType.STRUCT, 'uuid', (concrete.uuid.ttypes.UUID, concrete.uuid.ttypes.UUID.thrift_spec), None, ),  # 1
        (2, TType.STRUCT, 'metadata', (concrete.metadata.ttypes.AnnotationMetadata, concrete.metadata.ttypes.AnnotationMetadata.thrift_spec), None, ),  # 2
        (3, TType.LIST, 'situationList', (TType.STRUCT, (Situation, Situation.thrift_spec), False), None, ),  # 3
        (4, TType.LIST, 'linkingList', (TType.STRUCT, (concrete.linking.ttypes.Linking, concrete.linking.ttypes.Linking.thrift_spec), False), None, ),  # 4
    )

    def __init__(self, uuid=None, metadata=None, situationList=None, linkingList=None,):
        self.uuid = uuid
        self.metadata = metadata
        self.situationList = situationList
        self.linkingList = linkingList

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.uuid = concrete.uuid.ttypes.UUID()
                    self.uuid.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRUCT:
                    self.metadata = concrete.metadata.ttypes.AnnotationMetadata()
                    self.metadata.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.LIST:
                    self.situationList = []
                    (_etype38, _size35) = iprot.readListBegin()
                    for _i39 in range(_size35):
                        _elem40 = Situation()
                        _elem40.read(iprot)
                        self.situationList.append(_elem40)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.LIST:
                    self.linkingList = []
                    (_etype44, _size41) = iprot.readListBegin()
                    for _i45 in range(_size41):
                        _elem46 = concrete.linking.ttypes.Linking()
                        _elem46.read(iprot)
                        self.linkingList.append(_elem46)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('SituationSet')
        if self.uuid is not None:
            oprot.writeFieldBegin('uuid', TType.STRUCT, 1)
            self.uuid.write(oprot)
            oprot.writeFieldEnd()
        if self.metadata is not None:
            oprot.writeFieldBegin('metadata', TType.STRUCT, 2)
            self.metadata.write(oprot)
            oprot.writeFieldEnd()
        if self.situationList is not None:
            oprot.writeFieldBegin('situationList', TType.LIST, 3)
            oprot.writeListBegin(TType.STRUCT, len(self.situationList))
            for iter47 in self.situationList:
                iter47.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.linkingList is not None:
            oprot.writeFieldBegin('linkingList', TType.LIST, 4)
            oprot.writeListBegin(TType.STRUCT, len(self.linkingList))
            for iter48 in self.linkingList:
                iter48.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.uuid is None:
            raise TProtocolException(message='Required field uuid is unset!')
        if self.metadata is None:
            raise TProtocolException(message='Required field metadata is unset!')
        if self.situationList is None:
            raise TProtocolException(message='Required field situationList is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class MentionArgument(object):
    """
    A "concrete" argument, that may be used by SituationMentions or EntityMentions
    to avoid conflicts where abstract Arguments were being used to support concrete Mentions.

    Attributes:
     - role: The relationship between this argument and the situation that
    owns it. The roles that a situation's arguments can take
    depend on the type of the situation (including subtype
    information, such as event_type).
     - entityMentionId: A pointer to the value of an EntityMention, if this is being used to support
    an EntityMention.
     - situationMentionId: A pointer to the value of this argument, if it is a SituationMention.
     - tokens: The location of this MentionArgument in the Communication.
    If this MentionArgument can be identified in a document using an
    EntityMention or SituationMention, then UUID references to those
    types should be preferred and this field left as null.
     - constituent: An alternative way to specify the same thing as tokens.
     - confidence: Confidence of this argument belonging to its SituationMention
     - propertyList: For the BinarySRL task, there may be situations
    where more than one property is attached to a single
    participant. A list of these properties can be stored in this field.
    """

    thrift_spec = (
        None,  # 0
        (1, TType.STRING, 'role', 'UTF8', None, ),  # 1
        (2, TType.STRUCT, 'entityMentionId', (concrete.uuid.ttypes.UUID, concrete.uuid.ttypes.UUID.thrift_spec), None, ),  # 2
        (3, TType.STRUCT, 'situationMentionId', (concrete.uuid.ttypes.UUID, concrete.uuid.ttypes.UUID.thrift_spec), None, ),  # 3
        (4, TType.STRUCT, 'tokens', (concrete.structure.ttypes.TokenRefSequence, concrete.structure.ttypes.TokenRefSequence.thrift_spec), None, ),  # 4
        (5, TType.DOUBLE, 'confidence', None, None, ),  # 5
        (6, TType.LIST, 'propertyList', (TType.STRUCT, (Property, Property.thrift_spec), False), None, ),  # 6
        (7, TType.STRUCT, 'constituent', (concrete.structure.ttypes.ConstituentRef, concrete.structure.ttypes.ConstituentRef.thrift_spec), None, ),  # 7
    )

    def __init__(self, role=None, entityMentionId=None, situationMentionId=None, tokens=None, constituent=None, confidence=None, propertyList=None,):
        self.role = role
        self.entityMentionId = entityMentionId
        self.situationMentionId = situationMentionId
        self.tokens = tokens
        self.constituent = constituent
        self.confidence = confidence
        self.propertyList = propertyList

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.role = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRUCT:
                    self.entityMentionId = concrete.uuid.ttypes.UUID()
                    self.entityMentionId.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRUCT:
                    self.situationMentionId = concrete.uuid.ttypes.UUID()
                    self.situationMentionId.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.STRUCT:
                    self.tokens = concrete.structure.ttypes.TokenRefSequence()
                    self.tokens.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 7:
                if ftype == TType.STRUCT:
                    self.constituent = concrete.structure.ttypes.ConstituentRef()
                    self.constituent.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.DOUBLE:
                    self.confidence = iprot.readDouble()
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.LIST:
                    self.propertyList = []
                    (_etype52, _size49) = iprot.readListBegin()
                    for _i53 in range(_size49):
                        _elem54 = Property()
                        _elem54.read(iprot)
                        self.propertyList.append(_elem54)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('MentionArgument')
        if self.role is not None:
            oprot.writeFieldBegin('role', TType.STRING, 1)
            oprot.writeString(self.role.encode('utf-8') if sys.version_info[0] == 2 else self.role)
            oprot.writeFieldEnd()
        if self.entityMentionId is not None:
            oprot.writeFieldBegin('entityMentionId', TType.STRUCT, 2)
            self.entityMentionId.write(oprot)
            oprot.writeFieldEnd()
        if self.situationMentionId is not None:
            oprot.writeFieldBegin('situationMentionId', TType.STRUCT, 3)
            self.situationMentionId.write(oprot)
            oprot.writeFieldEnd()
        if self.tokens is not None:
            oprot.writeFieldBegin('tokens', TType.STRUCT, 4)
            self.tokens.write(oprot)
            oprot.writeFieldEnd()
        if self.confidence is not None:
            oprot.writeFieldBegin('confidence', TType.DOUBLE, 5)
            oprot.writeDouble(self.confidence)
            oprot.writeFieldEnd()
        if self.propertyList is not None:
            oprot.writeFieldBegin('propertyList', TType.LIST, 6)
            oprot.writeListBegin(TType.STRUCT, len(self.propertyList))
            for iter55 in self.propertyList:
                iter55.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.constituent is not None:
            oprot.writeFieldBegin('constituent', TType.STRUCT, 7)
            self.constituent.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class SituationMention(object):
    """
    A concrete mention of a situation, where "situations" include
    events, relations, facts, sentiments, and beliefs. Each situation
    has a core type (such as EVENT or SENTIMENT), along with an
    optional subtype based on its core type (e.g.,
    event_type=CONTACT_MEET), and a set of zero or more unordered
    arguments.

    This struct should be used for most types of SRL labelings
    (e.g. Propbank and FrameNet) because they are grounded in text.

    Attributes:
     - uuid: Unique identifier for this situation.
     - text: The text content of this situation mention. This field is
    often redundant with the 'tokens' field, and may not
    be generated by all analytics.
     - situationType: The core type of this situation (eg EVENT or SENTIMENT),
    or a coarse grain situation type.
     - situationKind: A fine grain situation type that specifically describes the
    situation mention based on situationType above. It allows for
    more detailed description of the situation mention.

    Some examples:

    if situationType == EVENT, the event type for the sit. mention
    if situationType == STATE, the state type for this sit. mention

    For Propbank, this field should be the predicate lemma and id,
    e.g. "strike.02". For FrameNet, this should be the frame name,
    e.g. "Commerce_buy".

    Different and more varied situationTypes may be added
    in the future.
     - argumentList: The arguments for this situation mention. Each argument
    consists of a role and a value. It is possible for an situation
    to have multiple arguments with the same role. Arguments are
    unordered.
     - intensity: An "intensity" rating for the situation, typically ranging from
    0-1. In the case of SENTIMENT situations, this is used to record
    the intensity of the sentiment.
     - polarity: The polarity of this situation. In the case of SENTIMENT
    situations, this is used to record the polarity of the
    sentiment.
     - tokens: An optional pointer to tokens that are (especially)
    relevant to this situation mention. It is left up to individual
    analytics to decide what tokens (if any) they wish to include in
    this field. In particular, it is not specified whether the
    arguments' tokens should be included.
     - constituent: An alternative way to specify the same thing as tokens.
     - confidence: A confidence score for this individual situation mention. You
    can also set a confidence score for an entire SituationMentionSet
    using the SituationMentionSet's metadata.
    """

    thrift_spec = (
        None,  # 0
        (1, TType.STRUCT, 'uuid', (concrete.uuid.ttypes.UUID, concrete.uuid.ttypes.UUID.thrift_spec), None, ),  # 1
        (2, TType.STRING, 'text', 'UTF8', None, ),  # 2
        (3, TType.STRING, 'situationType', 'UTF8', None, ),  # 3
        (4, TType.LIST, 'argumentList', (TType.STRUCT, (MentionArgument, MentionArgument.thrift_spec), False), None, ),  # 4
        None,  # 5
        None,  # 6
        None,  # 7
        None,  # 8
        None,  # 9
        None,  # 10
        None,  # 11
        None,  # 12
        None,  # 13
        None,  # 14
        None,  # 15
        None,  # 16
        None,  # 17
        None,  # 18
        None,  # 19
        None,  # 20
        None,  # 21
        None,  # 22
        None,  # 23
        None,  # 24
        None,  # 25
        None,  # 26
        None,  # 27
        None,  # 28
        None,  # 29
        None,  # 30
        None,  # 31
        None,  # 32
        None,  # 33
        None,  # 34
        None,  # 35
        None,  # 36
        None,  # 37
        None,  # 38
        None,  # 39
        None,  # 40
        None,  # 41
        None,  # 42
        None,  # 43
        None,  # 44
        None,  # 45
        None,  # 46
        None,  # 47
        None,  # 48
        None,  # 49
        (50, TType.STRING, 'situationKind', 'UTF8', None, ),  # 50
        None,  # 51
        None,  # 52
        None,  # 53
        None,  # 54
        None,  # 55
        None,  # 56
        None,  # 57
        None,  # 58
        None,  # 59
        None,  # 60
        None,  # 61
        None,  # 62
        None,  # 63
        None,  # 64
        None,  # 65
        None,  # 66
        None,  # 67
        None,  # 68
        None,  # 69
        None,  # 70
        None,  # 71
        None,  # 72
        None,  # 73
        None,  # 74
        None,  # 75
        None,  # 76
        None,  # 77
        None,  # 78
        None,  # 79
        None,  # 80
        None,  # 81
        None,  # 82
        None,  # 83
        None,  # 84
        None,  # 85
        None,  # 86
        None,  # 87
        None,  # 88
        None,  # 89
        None,  # 90
        None,  # 91
        None,  # 92
        None,  # 93
        None,  # 94
        None,  # 95
        None,  # 96
        None,  # 97
        None,  # 98
        None,  # 99
        (100, TType.DOUBLE, 'intensity', None, None, ),  # 100
        (101, TType.STRING, 'polarity', 'UTF8', None, ),  # 101
        None,  # 102
        None,  # 103
        None,  # 104
        None,  # 105
        None,  # 106
        None,  # 107
        None,  # 108
        None,  # 109
        None,  # 110
        None,  # 111
        None,  # 112
        None,  # 113
        None,  # 114
        None,  # 115
        None,  # 116
        None,  # 117
        None,  # 118
        None,  # 119
        None,  # 120
        None,  # 121
        None,  # 122
        None,  # 123
        None,  # 124
        None,  # 125
        None,  # 126
        None,  # 127
        None,  # 128
        None,  # 129
        None,  # 130
        None,  # 131
        None,  # 132
        None,  # 133
        None,  # 134
        None,  # 135
        None,  # 136
        None,  # 137
        None,  # 138
        None,  # 139
        None,  # 140
        None,  # 141
        None,  # 142
        None,  # 143
        None,  # 144
        None,  # 145
        None,  # 146
        None,  # 147
        None,  # 148
        None,  # 149
        (150, TType.STRUCT, 'tokens', (concrete.structure.ttypes.TokenRefSequence, concrete.structure.ttypes.TokenRefSequence.thrift_spec), None, ),  # 150
        (151, TType.STRUCT, 'constituent', (concrete.structure.ttypes.ConstituentRef, concrete.structure.ttypes.ConstituentRef.thrift_spec), None, ),  # 151
        None,  # 152
        None,  # 153
        None,  # 154
        None,  # 155
        None,  # 156
        None,  # 157
        None,  # 158
        None,  # 159
        None,  # 160
        None,  # 161
        None,  # 162
        None,  # 163
        None,  # 164
        None,  # 165
        None,  # 166
        None,  # 167
        None,  # 168
        None,  # 169
        None,  # 170
        None,  # 171
        None,  # 172
        None,  # 173
        None,  # 174
        None,  # 175
        None,  # 176
        None,  # 177
        None,  # 178
        None,  # 179
        None,  # 180
        None,  # 181
        None,  # 182
        None,  # 183
        None,  # 184
        None,  # 185
        None,  # 186
        None,  # 187
        None,  # 188
        None,  # 189
        None,  # 190
        None,  # 191
        None,  # 192
        None,  # 193
        None,  # 194
        None,  # 195
        None,  # 196
        None,  # 197
        None,  # 198
        None,  # 199
        (200, TType.DOUBLE, 'confidence', None, None, ),  # 200
    )

    def __init__(self, uuid=None, text=None, situationType=None, situationKind=None, argumentList=None, intensity=None, polarity=None, tokens=None, constituent=None, confidence=None,):
        self.uuid = uuid
        self.text = text
        self.situationType = situationType
        self.situationKind = situationKind
        self.argumentList = argumentList
        self.intensity = intensity
        self.polarity = polarity
        self.tokens = tokens
        self.constituent = constituent
        self.confidence = confidence

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.uuid = concrete.uuid.ttypes.UUID()
                    self.uuid.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.text = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.situationType = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 50:
                if ftype == TType.STRING:
                    self.situationKind = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.LIST:
                    self.argumentList = []
                    (_etype59, _size56) = iprot.readListBegin()
                    for _i60 in range(_size56):
                        _elem61 = MentionArgument()
                        _elem61.read(iprot)
                        self.argumentList.append(_elem61)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 100:
                if ftype == TType.DOUBLE:
                    self.intensity = iprot.readDouble()
                else:
                    iprot.skip(ftype)
            elif fid == 101:
                if ftype == TType.STRING:
                    self.polarity = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 150:
                if ftype == TType.STRUCT:
                    self.tokens = concrete.structure.ttypes.TokenRefSequence()
                    self.tokens.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 151:
                if ftype == TType.STRUCT:
                    self.constituent = concrete.structure.ttypes.ConstituentRef()
                    self.constituent.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 200:
                if ftype == TType.DOUBLE:
                    self.confidence = iprot.readDouble()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('SituationMention')
        if self.uuid is not None:
            oprot.writeFieldBegin('uuid', TType.STRUCT, 1)
            self.uuid.write(oprot)
            oprot.writeFieldEnd()
        if self.text is not None:
            oprot.writeFieldBegin('text', TType.STRING, 2)
            oprot.writeString(self.text.encode('utf-8') if sys.version_info[0] == 2 else self.text)
            oprot.writeFieldEnd()
        if self.situationType is not None:
            oprot.writeFieldBegin('situationType', TType.STRING, 3)
            oprot.writeString(self.situationType.encode('utf-8') if sys.version_info[0] == 2 else self.situationType)
            oprot.writeFieldEnd()
        if self.argumentList is not None:
            oprot.writeFieldBegin('argumentList', TType.LIST, 4)
            oprot.writeListBegin(TType.STRUCT, len(self.argumentList))
            for iter62 in self.argumentList:
                iter62.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.situationKind is not None:
            oprot.writeFieldBegin('situationKind', TType.STRING, 50)
            oprot.writeString(self.situationKind.encode('utf-8') if sys.version_info[0] == 2 else self.situationKind)
            oprot.writeFieldEnd()
        if self.intensity is not None:
            oprot.writeFieldBegin('intensity', TType.DOUBLE, 100)
            oprot.writeDouble(self.intensity)
            oprot.writeFieldEnd()
        if self.polarity is not None:
            oprot.writeFieldBegin('polarity', TType.STRING, 101)
            oprot.writeString(self.polarity.encode('utf-8') if sys.version_info[0] == 2 else self.polarity)
            oprot.writeFieldEnd()
        if self.tokens is not None:
            oprot.writeFieldBegin('tokens', TType.STRUCT, 150)
            self.tokens.write(oprot)
            oprot.writeFieldEnd()
        if self.constituent is not None:
            oprot.writeFieldBegin('constituent', TType.STRUCT, 151)
            self.constituent.write(oprot)
            oprot.writeFieldEnd()
        if self.confidence is not None:
            oprot.writeFieldBegin('confidence', TType.DOUBLE, 200)
            oprot.writeDouble(self.confidence)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.uuid is None:
            raise TProtocolException(message='Required field uuid is unset!')
        if self.argumentList is None:
            raise TProtocolException(message='Required field argumentList is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class SituationMentionSet(object):
    """
    A theory about the set of situation mentions that are present in a
    message. See also: SituationMention

    Attributes:
     - uuid: Unique identifier for this set.
     - metadata: Information about where this set came from.
     - mentionList: List of mentions in this set.
     - linkingList: Entity linking annotations associated with this SituationMentionSet.
    """

    thrift_spec = (
        None,  # 0
        (1, TType.STRUCT, 'uuid', (concrete.uuid.ttypes.UUID, concrete.uuid.ttypes.UUID.thrift_spec), None, ),  # 1
        (2, TType.STRUCT, 'metadata', (concrete.metadata.ttypes.AnnotationMetadata, concrete.metadata.ttypes.AnnotationMetadata.thrift_spec), None, ),  # 2
        (3, TType.LIST, 'mentionList', (TType.STRUCT, (SituationMention, SituationMention.thrift_spec), False), None, ),  # 3
        (4, TType.LIST, 'linkingList', (TType.STRUCT, (concrete.linking.ttypes.Linking, concrete.linking.ttypes.Linking.thrift_spec), False), None, ),  # 4
    )

    def __init__(self, uuid=None, metadata=None, mentionList=None, linkingList=None,):
        self.uuid = uuid
        self.metadata = metadata
        self.mentionList = mentionList
        self.linkingList = linkingList

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.uuid = concrete.uuid.ttypes.UUID()
                    self.uuid.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRUCT:
                    self.metadata = concrete.metadata.ttypes.AnnotationMetadata()
                    self.metadata.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.LIST:
                    self.mentionList = []
                    (_etype66, _size63) = iprot.readListBegin()
                    for _i67 in range(_size63):
                        _elem68 = SituationMention()
                        _elem68.read(iprot)
                        self.mentionList.append(_elem68)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.LIST:
                    self.linkingList = []
                    (_etype72, _size69) = iprot.readListBegin()
                    for _i73 in range(_size69):
                        _elem74 = concrete.linking.ttypes.Linking()
                        _elem74.read(iprot)
                        self.linkingList.append(_elem74)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('SituationMentionSet')
        if self.uuid is not None:
            oprot.writeFieldBegin('uuid', TType.STRUCT, 1)
            self.uuid.write(oprot)
            oprot.writeFieldEnd()
        if self.metadata is not None:
            oprot.writeFieldBegin('metadata', TType.STRUCT, 2)
            self.metadata.write(oprot)
            oprot.writeFieldEnd()
        if self.mentionList is not None:
            oprot.writeFieldBegin('mentionList', TType.LIST, 3)
            oprot.writeListBegin(TType.STRUCT, len(self.mentionList))
            for iter75 in self.mentionList:
                iter75.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.linkingList is not None:
            oprot.writeFieldBegin('linkingList', TType.LIST, 4)
            oprot.writeListBegin(TType.STRUCT, len(self.linkingList))
            for iter76 in self.linkingList:
                iter76.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.uuid is None:
            raise TProtocolException(message='Required field uuid is unset!')
        if self.metadata is None:
            raise TProtocolException(message='Required field metadata is unset!')
        if self.mentionList is None:
            raise TProtocolException(message='Required field mentionList is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
